﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.IO;
using System.IO.Pipes;
using System.Net.Sockets;
using System.Net;


namespace FileServant{
    public class Control
    {

        //these members are initialized by RunServer if necessary
        private FileStore fileStore = null;
        private List<FileStore.RootShare> shares = null;
        private HttpServer httpServerIPv4 = null;
        private HttpServer httpServerIPv6 = null;
        private FtpServer ftpServerIPv4 = null;
        private FtpServer ftpServerIPv6 = null;

        private volatile bool serverIsRunning = false;

        public FileStore.RootShare CreateShare(FileStore.URL url, String path, FileStore.Permission permission, String user, String password)
        {
            EnsureRunningServer();
            user = user == null ? "\"\"" : "\"" + user + "\"";
            password = password == null ? "\"\"" : "\"" + password + "\"";
            String strPermission = permission == FileStore.Permission.readwrite ? "rw" : "r";
            String result = RemoteExecute("CreateShare " + ToBase64(url.GetURL()) + " " + ToBase64(path) + " " + ToBase64(strPermission) + " " + ToBase64(user) + " " + ToBase64(password));
            if (result == "ERROR")
                return null; //failed to create share
            return FileStore.RootShare.FromString(result);
        }

        public String DeleteShare(FileStore.URL rootUrl)
        {
            EnsureRunningServer();
            String result = RemoteExecute("DeleteShare " + ToBase64(rootUrl.GetURL()));
            if (result == "ERROR") //failed to delete share
                return null;
            return result;
        }

        public String DeleteAll()
        {
            EnsureRunningServer();
            String result = RemoteExecute("DeleteAll");
            if (result == "ERROR")
                return null;
            return result;
        }

        public List<FileStore.RootShare> ListShares()
        {
            EnsureRunningServer();
            String result = RemoteExecute("ListShares");
            List<FileStore.RootShare> myShares = new List<FileStore.RootShare>();
            if (result == "ERROR") //failed to get list
                return myShares;
            if (result == "") //no shares exist
                return myShares;
            foreach (String s in result.Split(" ".ToCharArray()))
            {
                FileStore.RootShare share = FileStore.RootShare.FromString(FromBase64(s));
                if (share != null)
                    myShares.Add(share);
            }
            return myShares;
        }

        public int[] GetFtpHttpPorts()
        {
            EnsureRunningServer();
            int[] ports = new int[2];
            String result = RemoteExecute("Ports");
            String[] s = result.Split(" ".ToCharArray());
            try
            {
                ports[0] = int.Parse(s[0]);
                ports[1] = int.Parse(s[1]);
            }
            catch (Exception) { }
            return ports;
        }

        private TcpListener[] GetTcpListeners(int[] tryPorts)
        {
            TcpListener listenerIPv4 = null;
            TcpListener listenerIPv6 = null;
            bool v4Support = Socket.OSSupportsIPv4;
            bool v6Support = Socket.OSSupportsIPv6;

            //get a listener bound to first available port in list of requested ports
            for (int i = 0; i < tryPorts.Length; i++)
            {
                try
                {
                    //listener = new TcpListener(IPAddress.Any, tryPorts[i]);
                    if (v4Support && v6Support) {
                        //get two listeners on same port
                        listenerIPv4 = new TcpListener(IPAddress.Any, tryPorts[i]);
                        listenerIPv4.Start();
                        listenerIPv6 = new TcpListener(IPAddress.IPv6Any, ((IPEndPoint)listenerIPv4.LocalEndpoint).Port);
                        listenerIPv6.Start();
                    }
                    else if (v4Support)
                    {
                        listenerIPv4 = new TcpListener(IPAddress.Any, tryPorts[i]);
                        listenerIPv4.Start();
                    }
                    else if (v6Support)
                    {
                        listenerIPv6 = new TcpListener(IPAddress.IPv6Any, tryPorts[i]);
                        listenerIPv6.Start();
                    }
                }
                catch (Exception) {
                    try
                    {
                        if (listenerIPv4 != null)
                            listenerIPv4.Stop();
                    }
                    catch (Exception) { }
                    listenerIPv4 = null;
                    try
                    {
                        if (listenerIPv6 != null)
                            listenerIPv6.Stop();
                    }
                    catch (Exception) {  }
                    listenerIPv6 = null;
                }
                if (listenerIPv4 != null || listenerIPv6 != null)
                    break;
            }
            //none of requested ports were available - try to get listener for random port
            if (listenerIPv4 == null && listenerIPv6 == null)
            {
                int tries = 20;
                while (tries-- > 0)
                {
                    try
                    {
                        //listener = new TcpListener(IPAddress.Any, tryPorts[i]);
                        if (v4Support && v6Support)
                        {
                            //get two listeners on same port
                            listenerIPv4 = new TcpListener(IPAddress.Any, 0);
                            listenerIPv4.Start();
                            listenerIPv6 = new TcpListener(IPAddress.IPv6Any, ((IPEndPoint)listenerIPv4.LocalEndpoint).Port);
                            listenerIPv6.Start();
                        }
                        else if (v4Support)
                        {
                            listenerIPv4 = new TcpListener(IPAddress.Any, 0);
                            listenerIPv4.Start();
                        }
                        else if (v6Support)
                        {
                            listenerIPv6 = new TcpListener(IPAddress.IPv6Any, 0);
                            listenerIPv6.Start();
                        }
                    }
                    catch (Exception)
                    {
                        try
                        {
                            if (listenerIPv4 != null)
                                listenerIPv4.Stop();
                        }
                        catch (Exception) { }
                        listenerIPv4 = null;
                        try
                        {
                            if (listenerIPv6 != null)
                                listenerIPv6.Stop();
                        }
                        catch (Exception) { }
                        listenerIPv6 = null;
                    }
                    if (listenerIPv4 != null || listenerIPv6 != null)
                        break;
                }
            }
            return new TcpListener[] { listenerIPv4, listenerIPv6 };
        }


        private String Execute(String command)
        {
            try {
                String[] args = command.Split(" ".ToCharArray());
                for (int i = 1; i < args.Length; i++) {
                    args[i] = FromBase64(args[i]);
                }
                if (command.StartsWith("CreateShare "))
                {
                    String url = args[1];
                    String path = args[2];
                    String permission = args[3];
                    String user = args[4] == "\"\"" ? null : args[4].Substring(1, args[4].Length-2);
                    String password = args[5] == "\"\"" ? null : args[5].Substring(1, args[5].Length - 2);
                    FileStore.RootShare newShare = fileStore.CreateShare(new FileStore.URL(url), path, permission == "rw" ? FileStore.Permission.readwrite : FileStore.Permission.read, user, password);
                    if (newShare != null)
                    {
                        if (ftpServerIPv4 == null && ftpServerIPv6 == null)
                        {
                            TcpListener[] listeners = GetTcpListeners(new int[]{21, 2121, 2100, 21000, 21021, 21210});
                            TcpListener listerIPv4 = listeners[0];
                            TcpListener listerIPv6 = listeners[1];
                            if (listerIPv4 != null)
                                ftpServerIPv4 = new FtpServer(listerIPv4, new FtpFileHandler(fileStore));
                            if (listerIPv6 != null)
                                ftpServerIPv6 = new FtpServer(listerIPv6, new FtpFileHandler(fileStore));
                        }
                        if (httpServerIPv4 == null && httpServerIPv6 == null)
                        {
                            TcpListener[] listeners = GetTcpListeners(new int[] { 80, 8080, 8000, 8888, 8008, 8800, 8880, 800});
                            TcpListener listerIPv4 = listeners[0];
                            TcpListener listerIPv6 = listeners[1];
                            if (listerIPv4 != null)
                                httpServerIPv4 = new HttpServer(listerIPv4, new HttpFileHandler(fileStore));
                            if (listerIPv6 != null)
                                httpServerIPv6 = new HttpServer(listerIPv6, new HttpFileHandler(fileStore));
                        }

                        int insertAt = 0;
                        for (int i = 0; i < shares.Count; i++)
                        {
                            if (shares[i].rootDirName.CompareTo(url) > 0)
                            {
                                insertAt = i;
                                break;
                            }
                        }
                        shares.Insert(insertAt, newShare);

                        return newShare.ToString();
                    }
                }
                else if (command.StartsWith("DeleteShare "))
                {
                    Console.WriteLine("DELETESHARE!!!");
                    FileStore.URL rootShareUrl = new FileStore.URL(args[1]);
                    for (int i = 0; i < shares.Count; i++)
                    {
                        FileStore.RootShare share = shares[i];
                        Console.WriteLine(share.rootDirName);
                        Console.WriteLine(rootShareUrl.GetURL());
                        if (share.rootDirName == rootShareUrl.GetURL())
                        {
                            shares.Remove(share);
                            fileStore.DeleteShare(share);
                            if (shares.Count == 0)
                            {
                                if (ftpServerIPv4 != null)
                                {
                                    ftpServerIPv4.KillServer();
                                    ftpServerIPv4 = null;
                                }
                                if (ftpServerIPv6 != null)
                                {
                                    ftpServerIPv6.KillServer();
                                    ftpServerIPv6 = null;
                                }
                                if (httpServerIPv4 != null)
                                {
                                    httpServerIPv4.KillServer();
                                    httpServerIPv4 = null;
                                }
                                if (httpServerIPv6 != null)
                                {
                                    httpServerIPv6.KillServer();
                                    httpServerIPv6 = null;
                                }
                            }
                            return "SUCCESS";
                        }
                    }
                }
                else if (command == "DeleteAll")
                {
                    if (shares.Count > 0)
                    {
                        foreach (FileStore.RootShare share in shares)
                        {
                            fileStore.DeleteShare(share);
                        }
                        if (ftpServerIPv4 != null)
                        {
                            ftpServerIPv4.KillServer();
                            ftpServerIPv4 = null;
                        }
                        if (ftpServerIPv6 != null)
                        {
                            ftpServerIPv6.KillServer();
                            ftpServerIPv6 = null;
                        }
                        if (httpServerIPv4 != null)
                        {
                            httpServerIPv4.KillServer();
                            httpServerIPv4 = null;
                        }
                        if (httpServerIPv6 != null)
                        {
                            httpServerIPv6.KillServer();
                            httpServerIPv6 = null;
                        }
                        shares.Clear();
                    }
                    return "SUCCESS";
                }
                else if (command.StartsWith("ListShares"))
                {
                    String result = "";
                    foreach (FileStore.RootShare share in shares)
                    {
                        String s = ToBase64(share.ToString());
                        if (result == "")
                            result = s;
                        else
                            result += " " + s;
                    }
                    return result;
                }
                else if (command == "Ports")
                {
                    FtpServer ftpServer = ftpServerIPv4 != null ? ftpServerIPv4 : ftpServerIPv6;
                    HttpServer httpServer = httpServerIPv4 != null ? httpServerIPv4 : httpServerIPv6;
                    String result = "" + (ftpServer != null ? ftpServer.port : 0) + " " + (httpServer != null ? httpServer.port : 0);
                    return result;
                }
            } catch (Exception e) {
                //something has failed..
            }
            return "ERROR";
        }

        public String RemoteExecute(String command)
        {
            NamedPipeClientStream pipeClient = new NamedPipeClientStream(Util.toolPipeName);
            pipeClient.Connect(1000);
            if (pipeClient.IsConnected)
            {
                try
                {
                    StreamWriter writer = new StreamWriter(pipeClient);
                    writer.WriteLine(command);
                    writer.Flush();
                    String result = ReadLineFromStream(pipeClient);
                    return result;
                }
                finally
                {
                    pipeClient.Close();
                }
                
            }
            return null;
        }

        private static String ToBase64(String s)
        {
            char[] chars = s.ToCharArray();
            byte[] bytes = new byte[chars.Length];
            for (int i = 0; i < chars.Length; i++)
                bytes[i] = (byte)chars[i];
            return Convert.ToBase64String(bytes);
        }

        private static String FromBase64(String s)
        {
            byte[] bytes = Convert.FromBase64String(s);
            String result = "";
            for (int i = 0; i < bytes.Length; i++)
                result += (char)bytes[i];
            return result;
        }

        private static String ReadLineFromStream(Stream readFrom)
        {
            String result = "";
            int byteRead = readFrom.ReadByte();
            while (byteRead >= 0)
            {
                char c = (char)byteRead;
                if (c == '\n')
                    break;
                if (c != '\r')
                    result += c;
                byteRead = readFrom.ReadByte();
            }
            return result;
        }

        public void WaitForNoServer()
        {
            Thread.Sleep(250);
            while (serverIsRunning)
            {
                Thread.Sleep(1000);
            }
        }

        private void EnsureRunningServer()
        {
            NamedPipeServerStream pipeServer = null;
            lock (this)
            {
                try
                {
                    pipeServer = new NamedPipeServerStream(Util.toolPipeName, PipeDirection.InOut, 1, PipeTransmissionMode.Byte, PipeOptions.Asynchronous);
                }
                catch (Exception) { }
            }
            if (pipeServer != null)
            {
                Thread serverThread = new Thread(new ParameterizedThreadStart(this.RunServer));
                serverThread.Start(pipeServer);
            }
        }
        private void RunServer(Object o)
        {
            NamedPipeServerStream pipeServer = (NamedPipeServerStream)o;
            try
            {
                fileStore = new FileStore();
                shares = new List<FileStore.RootShare>();
                serverIsRunning = true;
                while (true)
                {
                    IAsyncResult pipeState = pipeServer.BeginWaitForConnection(PipeConnected, this);
                    if (pipeState.AsyncWaitHandle.WaitOne(10000))
                    {
                        //Client connected
                        pipeServer.EndWaitForConnection(pipeState);
                        try
                        {
                            String command = ReadLineFromStream(pipeServer);
                            String result = Execute(command);
                            StreamWriter writer = new StreamWriter(pipeServer);
                            writer.WriteLine(result);
                            writer.Flush();
                        }
                        catch (Exception ex)
                        {
                            //TBD: handle
                        }
                        finally
                        {
                            try
                            {
                                pipeServer.Disconnect();
                            }
                            catch (Exception) { }
                        }

                    }
                    else
                    {
                        //No client connected

                        //Close server if no shares exist
                        if (shares.Count == 0)
                            break;
                    }

                }
            }
            finally
            {
                pipeServer.Close();
                serverIsRunning = false;
            }
        }

        private void PipeConnected(IAsyncResult e)
        {
            //dummy method
        }

    }
}


