﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace FileServant
{
    public class FileStore
    {
        private List<RootShare> shares = new List<RootShare>();
        private Dictionary<String, RootShare> shareMap = new Dictionary<String, RootShare>();
        private enum CreateType { none, createFile, createDir };
        public enum Permission { read, readwrite };
        
        public RootShare CreateShare(URL shareURL, String fileSystemPath, Permission permission, String optionalUser, String optionalPassword)
        {
            try
            {
                lock (this)
                {
                    foreach (RootShare share in shares)
                    {
                        if (shareURL.GetURL() == share.rootDirName)
                            throw new FileStoreException("A share with this name already exists");
                    }
                    DirectoryInfo di = new DirectoryInfo(fileSystemPath);
                    
                    if (!di.Exists)
                        throw new FileStoreException("Path does not exist");
                    RootShare myShare = new RootShare(di, shareURL, permission, optionalUser, optionalPassword);
                    myShare.dirHandle = LockDirectory(di); //try prevent directory from being renamed or deleted
                    shares.Add(myShare);
                    shareMap.Add(shareURL.GetURL(), myShare);
                    Log.CreatedShare(myShare);
                    return myShare;
                }
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public void DeleteShare(RootShare share)
        {
            try
            {
                lock (this)
                {
                    if (!shares.Contains(share))
                        throw new FileStoreException("Share no int filestore");
                    share.online = false; //signal any threads serving files from share to abort
                    shares.Remove(share);
                    shareMap.Remove(share.rootDirName);
                    ReleaseDirectory(share.dirHandle); //close any open file handle for directory - release "lock"
                    Log.DeletedShare(share);
                }
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        //try to prevent directory from being renamed or deleted by "opening" a file handle
        private Microsoft.Win32.SafeHandles.SafeFileHandle LockDirectory(DirectoryInfo dir)
        {
            try
            {
                Microsoft.Win32.SafeHandles.SafeFileHandle dirHandle = Util.CreateFile(dir.FullName, Util.GENERIC_READ, Util.FILE_SHARE_READ | Util.FILE_SHARE_WRITE, IntPtr.Zero, Util.OPEN_EXISTING, Util.FILE_ATTRIBUTE_READONLY | Util.FILE_FLAG_BACKUP_SEMANTICS, IntPtr.Zero);
                if (!dirHandle.IsInvalid)
                    return dirHandle;
            }
            catch (Exception) { }
            return null;
        }

        private void ReleaseDirectory(Microsoft.Win32.SafeHandles.SafeFileHandle dirHandle)
        {
            try
            {
                if (dirHandle != null && !dirHandle.IsClosed)
                {
                    dirHandle.Close();
                }
            }
            catch (Exception) { }
        }


        private RootShare GetRootShareForPath(URL url) {
            String virtualPath = url.GetURL();
            //extract share name
            String rootDirPath = virtualPath;
            if (rootDirPath.IndexOf('/') != rootDirPath.LastIndexOf('/'))
            {
                rootDirPath = virtualPath.Substring(0, virtualPath.IndexOf('/', 1));
            }
            
            //look-up share
            RootShare rootShare = shareMap[rootDirPath];

            //check that share exists
            if (rootShare == null)
                throw new VirtualPathDoesNotExist(virtualPath);

            return rootShare;
        }

        //get or create file/dir and perform authentication control based on virtual root folder and user credentials
        private FileSystemInfo GetFileAndCheckPermissions(RootShare rootShare, URL url, CreateType createType, String user, String pass)
        {
            String virtualPath = url.GetURL();
            //check if access is permitted
            if (rootShare.user != null && rootShare.user != user || rootShare.pass != null && rootShare.pass != pass)
                throw new AuthenticationRequired(virtualPath);
            
            //construct corresponding filesystem path for virtualPath
            String myPath = rootShare.directory.FullName + "\\" + virtualPath.Substring(rootShare.rootDirName.Length).Replace('/', '\\');
            myPath = myPath.Replace("\\\\", "\\");

            DirectoryInfo parent = Directory.GetParent(myPath);
            bool exists = Directory.Exists(myPath);
            if ((parent == null && !Directory.Exists(myPath)) || (parent != null && !parent.Exists)) //ensure parent directory exists
                throw new VirtualPathDoesNotExist(virtualPath);
            if (createType == CreateType.none)
            {
                if (File.Exists(myPath))
                    return new FileInfo(myPath);
                if (Directory.Exists(myPath))
                    return new DirectoryInfo(myPath);
                throw new VirtualPathDoesNotExist(virtualPath);
            } else if (createType == CreateType.createFile && rootShare.permission == Permission.readwrite)
                return new FileInfo(myPath); //file will be created once .OpenWrite is called
            else if (createType == CreateType.createDir && rootShare.permission == Permission.readwrite)
            {
                DirectoryInfo di = new DirectoryInfo(myPath);
                di.Create();
                return di;
            }
            throw new FileStoreException("No write permission");
        }

        //Provide a list of files
        public List<VirtualFile> ListFiles(URL url, String user, String pass)
        {
            try
            {
                String virtualDirPath = url.GetURL();
                lock (this)
                {
                    if (virtualDirPath.Length <= 1) //retrieve list of shares
                    {
                        List<VirtualFile> files = new List<VirtualFile>();
                        foreach (RootShare share in shares)
                        {
                            VirtualFile vf = new VirtualFile(share.rootDirName, VirtualFile.filetype.directory, share.directory, share);
                            vf.permission = Permission.read;
                            files.Add(vf);
                        }
                        return files;
                    }
                    else //retrieve files/dirs from a share
                    {
                        RootShare rootShare = GetRootShareForPath(url);
                        DirectoryInfo di = (DirectoryInfo)GetFileAndCheckPermissions(rootShare, url, CreateType.none, user, pass);
                        List<VirtualFile> files = new List<VirtualFile>();
                        foreach (FileSystemInfo fsi in di.GetFileSystemInfos())
                        {
                            if ((fsi.Attributes & FileAttributes.Directory) == FileAttributes.Directory)
                                files.Add(new VirtualFile(virtualDirPath + "/" + fsi.Name, VirtualFile.filetype.directory, fsi, rootShare));
                            else
                                files.Add(new VirtualFile(virtualDirPath + "/" + fsi.Name, VirtualFile.filetype.file, fsi, rootShare));
                        }
                        return files;
                    }
                }
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public VirtualFile GetFile(URL url, String user, String pass, bool createFile)
        {
            try
            {
                String virtualFilePath = url.GetURL();
                lock (this)
                {
                    if (virtualFilePath == "/")
                    {
                        return new VirtualFile("/", VirtualFile.filetype.directory, null, null);
                    }
                    RootShare rootShare = GetRootShareForPath(url);
                    FileSystemInfo fi = (FileSystemInfo)GetFileAndCheckPermissions(rootShare, url, createFile ? CreateType.createFile : CreateType.none, user, pass);
                    return new VirtualFile(virtualFilePath, (fi is DirectoryInfo) ? VirtualFile.filetype.directory : VirtualFile.filetype.file, fi, rootShare);
                }
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public void Delete(VirtualFile file, String user, String pass)
        {
            try {
                if (file.share.user != null && file.share.user != user)
                    throw new AuthenticationRequired(file.virtualPath);
                if (file.share.pass != null && file.share.pass != pass)
                    throw new AuthenticationRequired(file.virtualPath);
                if (file.share.permission == Permission.readwrite)
                    file.theFile.Delete();
                else
                    throw new FileStoreException("File is read-only");
            } catch (Exception ex) {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public Stream ReadFile(VirtualFile file, String user, String pass)
        {
            try
            {
                if (file.share.user != null && file.share.user != user)
                    throw new AuthenticationRequired(file.virtualPath);
                if (file.share.pass != null && file.share.pass != pass)
                    throw new AuthenticationRequired(file.virtualPath);
                return ((FileInfo)file.theFile).Open(FileMode.Open, FileAccess.Read, FileShare.Read);
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public Stream WriteFile(VirtualFile file, long writeFromIndex, String user, String pass)
        {
            try
            {
                if (file.share.user != null && file.share.user != user)
                    throw new AuthenticationRequired(file.virtualPath);
                if (file.share.pass != null && file.share.pass != pass)
                    throw new AuthenticationRequired(file.virtualPath);
                long size = file.theFile.Exists ? file.GetSize() : 0;
                Stream fs = null;
                if (file.theFile.Exists)
                    fs = ((FileInfo)file.theFile).Open(FileMode.Open, FileAccess.Write, FileShare.Read);
                else 
                    fs = ((FileInfo)file.theFile).Create();
                if (writeFromIndex > size)
                    throw new FileStoreException("File size is less than resume index");
                if (writeFromIndex > 0)
                    fs.SetLength(writeFromIndex);
                fs.Seek(writeFromIndex, SeekOrigin.Begin);
                return fs;
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public void Rename(URL renameFrom, URL renameTo, String user, String pass)
        {
            FileStore.URL parentURL = new FileStore.URL(renameTo.GetURL().Substring(0, renameTo.GetURL().LastIndexOf('/')));
            if (renameFrom.GetURL().LastIndexOf('/') == 0 || renameTo.GetURL().LastIndexOf('/') == 0 )
            {
                throw new FileStoreException("Cannot rename root elements.");
            }
            try
            {
                FileStore.VirtualFile fromFile = GetFile(renameFrom, user, pass, false);
                if (fromFile.permission == FileStore.Permission.read)
                {
                    throw new FileStoreException("Read only");
                }
                FileStore.VirtualFile toFile = null;
                try
                {
                    toFile = GetFile(renameTo, user, pass, false); //check permissions and existence of target
                }
                catch (VirtualPathDoesNotExist) { }
                if (toFile != null) //error - target already exists
                {
                    throw new FileStoreException("Target already exists");
                }
                toFile = GetFile(renameTo, user, pass, true);
                //all checks passed - perform move
                if (fromFile.type == VirtualFile.filetype.directory) {
                    ((DirectoryInfo)fromFile.theFile).MoveTo(toFile.theFile.FullName);
                } else {
                    ((FileInfo)fromFile.theFile).MoveTo(toFile.theFile.FullName);
                }
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public VirtualFile CreateDirectory(URL url, String user, String pass)
        {
            try
            {
                String virtualDirPath = url.GetURL();
                lock (this)
                {
                    RootShare rootShare = GetRootShareForPath(url);
                    if (rootShare.permission == Permission.readwrite)
                    {
                        DirectoryInfo di = (DirectoryInfo)GetFileAndCheckPermissions(rootShare, url, CreateType.createDir, user, pass);
                        return new VirtualFile(virtualDirPath, VirtualFile.filetype.directory, di, rootShare);
                    }
                    else
                    {
                        throw new Exception("Target is read-only " + virtualDirPath);
                    }
                }
            }
            catch (Exception ex)
            {
                if (ex is FileStoreException)
                    throw ex;
                throw new FileStoreException(ex);
            }
        }

        public class URL
        {
            private String url;

            public URL(String argUrl)
            {
                try
                {
                    url = argUrl;
                    if (url == null)
                        throw new NullReferenceException();
                    //remove trailing /
                    while (url.Length > 1 && url.LastIndexOf('/') == url.Length - 1)
                        url = url.Substring(0, url.Length - 1);
                    //remove double //
                    while (url.IndexOf("//") >= 0)
                        url = url.Replace("//", "/");
                    while (url.IndexOf("/../") >= 0)
                        url = url.Replace("/../", "/");
                    if (url.EndsWith("/.."))
                        url = url.Replace("/..", "");
                    if (!url.StartsWith("/"))
                        url = "/" + url;
                    //strip some more illegal characters \:*?"<>|
                    url = url.Replace("\\", "").Replace(":", "").Replace(":", "").Replace("*", "").Replace("?", "").Replace("\"", "").Replace("<", "").Replace(">", "").Replace("|", "");
                }
                catch (Exception ex)
                {
                    if (ex is FileStoreException)
                        throw ex;
                    throw new FileStoreException(ex);
                }
            }

            public String GetURL()
            {
                return url;
            }
        }

        public class RootShare
        {
            public DirectoryInfo directory;
            public String rootDirName;
            public String user;
            public String pass;
            public Permission permission;
            public bool online = true; //primitive method to signal a cancel request to anyone testing this value
            public Microsoft.Win32.SafeHandles.SafeFileHandle dirHandle = null;
            
            public RootShare(DirectoryInfo argDirectory, URL argRootDirName, Permission argPermission, String argUser, String argPass)
            {
                directory = argDirectory;
                rootDirName = argRootDirName.GetURL();
                user = argUser;
                pass = argPass;
                permission = argPermission;
                //validate that rootDirName is on acceptable form of '/dirname'
                if (rootDirName.Length < 2 || rootDirName.LastIndexOf("/") != rootDirName.IndexOf("/"))
                    throw new Exception("Not a valid RootShare path " + rootDirName);
            }

            public override string ToString()
            {
                return rootDirName + "|" + directory.FullName + "|" + (permission == FileStore.Permission.readwrite ? "rw" : "r") + "|" + (user != null ? user : "") + "|" + (pass != null ? pass : "");
            }

            public static RootShare FromString(String str)
            {
                try
                {
                    String[] s = str.Split("|".ToCharArray());
                    if (s.Length != 5)
                        return null;
                    URL argURL = new URL(s[0]);
                    DirectoryInfo argDir = new DirectoryInfo(s[1]);
                    Permission argPerm = s[2] == "rw" ? Permission.readwrite : Permission.read;
                    String argUser = s[3];
                    if (argUser == "")
                        argUser = null;
                    String argPass = s[4];
                    if (argPass == "")
                        argPass = null;
                    return new RootShare(argDir, argURL, argPerm, argUser, argPass);
                }
                catch (Exception) { }
                return null;
            }
        }

        public class VirtualFile
        {
            public enum filetype { file, directory };
            public String virtualPath;
            public String name;
            public filetype type;
            public FileSystemInfo theFile;
            public RootShare share;
            public Permission permission;

            public VirtualFile(String argVirtualPath, filetype argType, FileSystemInfo argFile, RootShare argShare)
            {
                virtualPath = argVirtualPath;
                type = argType;
                theFile = argFile;
                share = argShare;
                permission = share != null ? share.permission : Permission.read;
                try
                {
                    name = virtualPath != "/" ? virtualPath.Substring(virtualPath.LastIndexOf('/')+1) : virtualPath.Substring(1);

                }
                catch (Exception)
                {
                }
            }

            public DateTime GetTime()
            {
                if (theFile != null)
                    return theFile.CreationTime;
                else
                    return new DateTime(1980, 11, 23);
            }

            public long GetSize()
            {
                if (theFile is FileInfo)
                {
                    return ((FileInfo)theFile).Length;
                }
                return 0;
            }

            public bool Exists()
            {
                return theFile.Exists;
            }

        }

        public class FileStoreException : Exception
        {
            public FileStoreException(Exception innerException)
                : base("General Exception", innerException)
            {
            }

            public FileStoreException(String message)
                : base(message)
            {
            }
        }

        public class VirtualPathDoesNotExist : FileStoreException
        {
            public VirtualPathDoesNotExist(String virtualPath)
                : base(virtualPath)
            {
            }
        }

        
        public class AuthenticationRequired : FileStoreException
        {
            public AuthenticationRequired(String virtualPath)
                : base(virtualPath)
            {
            }
        }
    }
}
