﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace FileServant
{
    public class FtpFileHandler : FtpServer.FtpHandler
    {
        private FileStore fileStore;
        private enum Type { ASCII, Binary };
        private static System.Globalization.CultureInfo enUS = new System.Globalization.CultureInfo("en-US");
        private static System.Net.IPAddress clientIP;

        public FtpFileHandler(FileStore argFileStore)
        {
            fileStore = argFileStore;
        }

        public void Handle(String user, String pass, Stream commandStream, FtpServer.DataStreamHandler dataStreamHandler, System.Net.IPAddress argClientIP)
        {
            clientIP = argClientIP;
            FileStore.VirtualFile cwd = fileStore.GetFile(new FileStore.URL("/"), user, pass, false);
            FileStore.RootShare cwdShare = null;
            FileStore.URL rnfr = null;
            StreamWriter response = Util.GetStreamWriter(commandStream);
            response.AutoFlush = true;
            long restartPosition = 0;

            Type type = Type.ASCII;
            while (cwdShare == null || cwdShare.online)
            {
                String command = Util.ReadLineFromStream(commandStream).Trim();
                String commandLow = command.ToLower();
                if (commandLow.StartsWith("user"))
                {
                    response.Write("530 Can't change to another user.\r\n");
                }
                else if (commandLow.StartsWith("quit"))
                {
                    response.Write("221 Goodbye.\r\n");
                    return;
                }
                else if (commandLow.StartsWith("port"))
                {
                    String portStr = command.Substring(command.IndexOf(' '));
                    try
                    {
                        dataStreamHandler.SetPORT(portStr);
                        response.Write("200 PORT command successful. Consider using PASV.\r\n");
                    }
                    catch (FtpServer.FtpException ftpException)
                    {
                        response.Write(ftpException.Message + "\r\n");
                    }
                }
                else if (commandLow.StartsWith("eprt"))
                {
                    String portStr = command.Substring(command.IndexOf(' '));
                    try
                    {
                        dataStreamHandler.SetEPRT(portStr);
                        response.Write("200 EPRT command successful. Consider using EPSV.\r\n");
                    }
                    catch (FtpServer.FtpException ftpException)
                    {
                        response.Write(ftpException.Message + "\r\n");
                    }
                }
                else if (commandLow.StartsWith("pasv"))
                {
                    if (commandLow == "pasv")
                    {
                        String result = dataStreamHandler.SetPASV();
                        response.Write("227 Entering Passive Mode " + result + ".\r\n");
                    }
                    else
                    {
                        response.Write("500 Syntax error, command unrecognized.\r\n");
                    }
                }
                else if (commandLow.StartsWith("epsv"))
                {
                    //TBD: handle EPSV ALL (see rfc 2428)
                    String arg = null;
                    if (commandLow.StartsWith("epsv "))
                        arg = command.Substring(command.IndexOf(' '));
                    String result = dataStreamHandler.SetEPSV(arg, clientIP.AddressFamily);
                    response.Write("229 Entering Extended Passive Mode " + result + ".\r\n");
                }
                else if (commandLow.StartsWith("type"))
                {
                    if (commandLow == "type a")
                    {
                        type = Type.ASCII;
                        response.Write("200 Switching to ASCII mode.\r\n");
                    }
                    else if (commandLow == "type i")
                    {
                        type = Type.Binary;
                        response.Write("200 Switching to Binary mode.\r\n");
                    }
                    else
                    {
                        response.Write("500 Unrecognised TYPE command.\r\n");
                    }
                }
                else if (commandLow.StartsWith("mode"))
                {
                    if (commandLow == "mode s")
                    {
                        response.Write("200 Mode set to STREAM.\r\n");
                    }
                    else
                    {
                        response.Write("504 Bad MODE command. Only S is supported!\r\n");
                    }
                }
                else if (commandLow.StartsWith("stru"))
                {
                    if (commandLow == "stru f")
                    {
                        response.Write("200 Structure set to FILE.\r\n");
                    }
                    else
                    {
                        response.Write("504 Bad STRU command. Only F is supported!\r\n");
                    }
                }
                else if (commandLow.StartsWith("noop"))
                {
                    if (commandLow == "noop")
                    {
                        response.Write("200 NOOP ok.\r\n");
                    }
                    else
                    {
                        response.Write("500 Syntax error, command unrecognized.\r\n");
                    }
                }
                else if (commandLow.StartsWith("pwd") || commandLow.StartsWith("xpwd"))
                {
                    if (commandLow == "pwd" || commandLow == "xpwd")
                    {
                        response.Write("257 \"" + cwd.virtualPath.Replace("\"", "\"\"") + "\"\r\n");
                    }
                    else
                    {
                        response.Write("500 Syntax error, command unrecognized.\r\n");
                    }
                }
                else if (commandLow.StartsWith("cwd") || commandLow.StartsWith("cdup"))
                {
                    try
                    {
                        String dir = command.Substring(command.IndexOf(' ') + 1);
                        if (commandLow.StartsWith("cdup") || dir == "..")
                            dir = cwd.virtualPath.Substring(0, cwd.virtualPath.LastIndexOf('/'));
                        else if (!dir.StartsWith("/"))
                            dir = cwd.virtualPath + "/" + dir;
                        FileStore.VirtualFile vf = fileStore.GetFile(new FileStore.URL(dir), user, pass, false);
                        if (vf.type == FileStore.VirtualFile.filetype.directory)
                        {
                            cwd = vf;
                            response.Write("250 Directory successfully changed.\r\n");
                        }
                        else
                        {
                            response.Write("550 Failed to change directory. Path is a file!\r\n");
                        }
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 Failed to change directory.\r\n");
                    }
                }
                else if (commandLow.StartsWith("mkd"))
                {
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.URL pathURL = new FileStore.URL(path);
                        FileStore.VirtualFile vf = fileStore.CreateDirectory(pathURL, user, pass);
                        Log.FileWrite(clientIP, vf);
                        response.Write("257 \"" + pathURL.GetURL().Replace("\"", "\"\"") + "\" created.\r\n");
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 Failed to make a directory.\r\n");
                    }
                }
                else if (commandLow.StartsWith("rmd"))
                {
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.URL pathURL = new FileStore.URL(path);

                        FileStore.VirtualFile vf = fileStore.GetFile(pathURL, user, pass, false);
                        if (vf.type == FileStore.VirtualFile.filetype.file || vf.permission == FileStore.Permission.read || vf.virtualPath.LastIndexOf('/') == 0)
                        {
                            response.Write("550 Failed to delete directory.\r\n");
                        }
                        else
                        {
                            Log.FileWrite(clientIP, vf);
                            fileStore.Delete(vf, user, pass);
                            response.Write("250 Directory deleted.\r\n");
                        }
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 Failed to delete directory.\r\n");
                    }
                }
                else if (commandLow.StartsWith("feat"))
                {
                    //list features not defined in rfc 959
                    response.Write("211-Features:\r\n");
                    response.Write(" UTF8\r\n");
                    response.Write(" XPWD\r\n");
                    response.Write(" EPSV\r\n");
                    response.Write(" EPRT\r\n");
                    response.Write("211 End\r\n");
                }
                else if (commandLow.StartsWith("opts"))
                {
                    if (commandLow == "opts utf8 on")
                    {
                        response.Write("200 Always in UTF8 mode.\r\n");
                    }
                    else
                    {
                        response.Write("501 Option not understood.\r\n");
                    }
                }
                else if (commandLow.StartsWith("rnfr"))
                {
                    rnfr = null;
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.URL fromURL = new FileStore.URL(path);
                        FileStore.VirtualFile vf = fileStore.GetFile(fromURL, user, pass, false);
                        if (vf.permission == FileStore.Permission.readwrite)
                        {
                            rnfr = fromURL;
                            response.Write("350 Ready for RNTO.\r\n");
                        }
                        else
                        {
                            response.Write("550 RNFR command failed.\r\n");
                        }
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 RNFR command failed.\r\n");
                    }
                }
                else if (commandLow.StartsWith("rnto"))
                {
                    if (rnfr == null)
                    {
                        response.Write("503 RNFR required first.\r\n");
                    }
                    else
                    {
                        try
                        {
                            String path = command.Substring(command.IndexOf(' ') + 1);
                            if (!path.StartsWith("/"))
                                path = cwd.virtualPath + "/" + path;
                            FileStore.URL toURL = new FileStore.URL(path);
                            FileStore.VirtualFile fromFile = fileStore.GetFile(rnfr, user, pass, false);
                            FileStore.VirtualFile toFile = fileStore.GetFile(toURL, user, pass, true);
                            fileStore.Rename(rnfr, toURL, user, pass);
                            Log.FileWrite(clientIP, fromFile);
                            Log.FileWrite(clientIP, toFile);
                            rnfr = null;
                            response.Write("250 Rename successful.\r\n");
                        }
                        catch (FileStore.AuthenticationRequired)
                        {
                            response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                        }
                        catch (Exception)
                        {
                            response.Write("550 Rename failed.\r\n");
                        }
                    }
                }
                else if (commandLow.StartsWith("size"))
                {
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.VirtualFile vf = fileStore.GetFile(new FileStore.URL(path), user, pass, false);
                        if (vf.type == FileStore.VirtualFile.filetype.file)
                        {
                            long size = vf.GetSize();
                            response.Write("213 " + size + "\r\n");
                        }
                        else
                        {
                            response.Write("550 Could not get file size.\r\n");
                        }
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 Could not get file size.\r\n");
                    }
                }
                else if (commandLow.StartsWith("dele"))
                {
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.URL pathURL = new FileStore.URL(path);

                        FileStore.VirtualFile vf = fileStore.GetFile(pathURL, user, pass, false);
                        Log.FileWrite(clientIP, vf);
                        fileStore.Delete(vf, user, pass);
                        response.Write("250 File deleted.\r\n");
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 Failed to delete file.\r\n");
                    }
                }
                else if (commandLow.StartsWith("rest"))
                {
                    restartPosition = 0;
                    try
                    {
                        restartPosition = long.Parse(commandLow.Substring(commandLow.IndexOf(' ') + 1));
                    }
                    catch (Exception)
                    {
                    }
                    if (restartPosition < 0)
                        restartPosition = 0;
                    response.Write("350 Restart position accepted (" + restartPosition + ").\r\n");
                }
                else if (commandLow.StartsWith("retr"))
                {
                    long index = restartPosition;
                    restartPosition = 0;

                    BufferedStream readStream = null;
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.VirtualFile vf = fileStore.GetFile(new FileStore.URL(path), user, pass, false);
                        long size = vf.GetSize();
                        if (vf.type == FileStore.VirtualFile.filetype.file)
                        {
                            Log.FileServe(clientIP, vf);

                            readStream = new BufferedStream(fileStore.ReadFile(vf, user, pass));

                            Stream dataStream = new BufferedStream(dataStreamHandler.GetDataStream());

                            response.Write("150 Opening BINARY mode data connection for " + vf.name + " (" + (size - index) + " bytes).\r\n");

                            if (index > 0) //fast-forward to required bytes
                                readStream.Seek(index, SeekOrigin.Begin);

                            byte[] buffer = new byte[524288]; //new byte[4096];
                            while (index < size && vf.share.online)
                            {
                                int read = readStream.Read(buffer, 0, buffer.Length);
                                if (read == 0)
                                    break; //break out of loop - end of file reached - if this was before byteTotal we dont care and fail silently
                                dataStream.Write(buffer, 0, read);
                                index += read;
                            }
                            dataStream.Flush();
                            response.Write("226 File send OK.\r\n");
                        }
                        else
                        {
                            response.Write("550 Failed to open file.\r\n");
                        }
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("550 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (FtpServer.FtpException ftpEx)
                    {
                        response.Write(ftpEx.Message + "\r\n");
                    }
                    catch (Exception)
                    {
                        response.Write("550 An error occured.\r\n");
                    }
                    finally
                    {
                        dataStreamHandler.CloseAnyClient();
                        if (readStream != null)
                            readStream.Close();
                    }
                }
                else if (commandLow.StartsWith("list") || commandLow.StartsWith("nlst") || commandLow.StartsWith("stat "))
                {
                    try
                    {
                        bool verbose = commandLow.StartsWith("list") || commandLow.StartsWith("stat");
                        bool useControlConnection = commandLow.StartsWith("stat");
                        if (useControlConnection)
                            response.Write("213-Status follows:\r\n");
                        else
                            response.Write("150 Here comes the directory listing.\r\n");
                        Stream output = useControlConnection ? null : dataStreamHandler.GetDataStream();

                        //strip any options from the command before extracting path - we ignore options and serve everything as if specified -la
                        //nb. this might be problematic if a user wishes to list a directory named '-a' - instead current directory would be listed
                        String[] tmp = command.Split(new char[]{' '});
                        command = "";
                        int stopStrip = 1; //number of substrings that does not start with a '-' to accept before stopping purge (protect from purging filenames that contain ' -'
                        foreach (string s in tmp) {
                            if (stopStrip < 0 || !s.StartsWith("-"))
                            {
                                command += " " + s;
                                stopStrip--;
                            }
                        }
                        command = command.Substring(1);

                        String path = cwd.virtualPath;
                        if (command.IndexOf(' ') > 0)
                            path = command.Substring(command.IndexOf(' ') + 1);
                        List<String> list = new List<string>();
                        try
                        {
                            FileStore.VirtualFile virtualDir = fileStore.GetFile(new FileStore.URL(path), user, pass, false);
                            Log.DirServe(clientIP, virtualDir);
                            foreach (FileStore.VirtualFile vf in fileStore.ListFiles(new FileStore.URL(path), user, pass))
                            {
                                //mode links owner group size datetime name
                                String mode = null;
                                if (vf.permission == FileStore.Permission.read)
                                {
                                    mode = vf.type == FileStore.VirtualFile.filetype.directory ? "dr-xr-xr-x" : "-r--r--r--";
                                }
                                else
                                {
                                    mode = vf.type == FileStore.VirtualFile.filetype.directory ? "drwxrwxrwx" : "-rw-rw-rw-";
                                }
                                String links = "  1";
                                String owner = (user + "        ").Substring(0, 10);
                                String group = user;
                                String size = vf.type == FileStore.VirtualFile.filetype.file ? "" + vf.GetSize() : "4096";
                                size = size.PadLeft(12, ' ');
                                DateTime fileDate = vf.GetTime();
                                String datetime = null;
                                if (DateTime.Now.Subtract(fileDate).Days > 180)
                                {
                                    datetime = fileDate.ToString("MMM dd  yyyy", enUS);
                                }
                                else
                                {
                                    datetime = fileDate.ToString("MMM dd HH:mm", enUS);
                                }
                                String name = vf.virtualPath.Substring(vf.virtualPath.LastIndexOf('/') + 1);
                                if (verbose)
                                    list.Add(mode + " " + links + " " + owner + " " + group + " " + size + " " + datetime + " " + name);
                                else
                                    list.Add(name);
                            }
                        }
                        catch (Exception)
                        {
                            //ignore and serve empty list
                            list.Clear();
                        }
                        if (useControlConnection)
                        {
                            foreach (String s in list)
                            {
                                response.Write(s + "\r\n");
                            }
                            response.Write("213 End of status\r\n");
                            response.Flush();
                        }
                        else
                        {
                            StreamWriter outputWriter = Util.GetStreamWriter(output);
                            foreach (String s in list)
                            {
                                outputWriter.Write(s + "\r\n");
                            }
                            outputWriter.Flush();
                            output.Flush();
                            response.Write("226 Directory send OK.\r\n");
                            output.Close();
                        }
                    }
                    catch (FtpServer.FtpException ftpException)
                    {
                        response.Write(ftpException.Message + "\r\n");
                    }
                    finally
                    {
                        dataStreamHandler.CloseAnyClient();
                    }
                }
                else if (commandLow.StartsWith("stor") || commandLow.StartsWith("appe"))
                {
                    long index = restartPosition;
                    restartPosition = 0;

                    BufferedStream writeStream = null;
                    try
                    {
                        String path = command.Substring(command.IndexOf(' ') + 1);
                        if (!path.StartsWith("/"))
                            path = cwd.virtualPath + "/" + path;
                        FileStore.URL targetPath = new FileStore.URL(path);
                        FileStore.URL parentURL = new FileStore.URL(targetPath.GetURL().Substring(0, targetPath.GetURL().LastIndexOf('/')));


                        FileStore.VirtualFile targetFile = fileStore.GetFile(targetPath, user, pass, true);



                        if (commandLow.StartsWith("appe"))
                        {
                            if (targetFile.Exists())
                                index = targetFile.GetSize();
                            else
                                index = 0;
                        }

                        Stream fileStream = fileStore.WriteFile(targetFile, index, user, pass);
                        writeStream = new BufferedStream(fileStream);

                        Log.FileWrite(clientIP, targetFile);

                        Stream dataStream = new BufferedStream(dataStreamHandler.GetDataStream());

                        response.Write("150 Opening BINARY mode data connection for " + targetFile.name + ".\r\n");

                        byte[] buffer = new byte[524288]; //new byte[4096];
                        long bytesRecieved = 0;
                        while (targetFile.share.online)
                        {
                            int read = dataStream.Read(buffer, 0, buffer.Length);
                            if (read == 0)
                                break; //break out of loop - end of file reached
                            writeStream.Write(buffer, 0, read);
                            bytesRecieved += read;
                        }

                        writeStream.Flush();

                        response.Write("226 File written OK (" + bytesRecieved + " bytes).\r\n");
                    }
                    catch (FileStore.AuthenticationRequired)
                    {
                        response.Write("532 Blocked. Log in with correct username and password to access!\r\n");
                    }
                    catch (FtpServer.FtpException ftpEx)
                    {
                        response.Write(ftpEx.Message + "\r\n");
                    }
                    catch (Exception ex)
                    {
                        response.Write("553 An error occured.\r\n");
                    }
                    finally
                    {
                        dataStreamHandler.CloseAnyClient();
                        if (writeStream != null)
                            writeStream.Close();
                    }
                }
                else if (commandLow.StartsWith("syst"))
                {
                    if (commandLow == "syst")
                        response.Write("215 UNIX Type: L8\r\n");
                    else
                        response.Write("500 Syntax error, command unrecognized.\r\n");
                }
                else
                {
                    response.Write("500 Unknown command.\r\n");
                }
            }
        }
    }
}
