﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using System.IO;
using System.Threading;
using System.Net;
using System.Web;

namespace FileServant
{
    public class FtpServer
    {
        private TcpListener connectionListener;
        private FtpHandler handler;
        private List<Session> pending = new List<Session>();
        private Thread serverThread;
        private bool running = true;
        public int port = 0;
        public AddressFamily addressFamily;

        public FtpServer(TcpListener argConnectionListener, FtpHandler argHandler)
        {
            connectionListener = argConnectionListener;
            handler = argHandler;
            if (handler == null)
                throw new NullReferenceException();
            port = ((System.Net.IPEndPoint)connectionListener.LocalEndpoint).Port;
            addressFamily = connectionListener.LocalEndpoint.AddressFamily;
            serverThread = new Thread(new ThreadStart(this.RunServer));
            serverThread.Start();
        }

        public void KillServer()
        {
            running = false;
            //stop accepting new connections and kill any helper threads
            try
            {
                connectionListener.Stop();
            }
            catch (Exception ignore)
            {
            }

            //wait until all pending requests has been cancelled before returning
            int numPending = 1;
            while (numPending > 0)
            {
                lock (this)
                {
                    numPending = pending.Count;
                }
                if (numPending == 0)
                    break;
                Thread.Sleep(100);
            }
        }

        public interface FtpHandler
        {
            void Handle(String user, String pass, Stream commandStream, DataStreamHandler dataStreamHandler, System.Net.IPAddress clientIP); 
        }

        private void Closed(Session session)
        {
            lock (this)
            {
                pending.Remove(session);
            }
        }

        private void RunServer()
        {
            Log.FtpOnline(this);
            while (running)
            {
                try
                {
                    TcpClient client = connectionListener.AcceptTcpClient();
                    lock (this)
                    {
                        pending.Add(new Session(this, client, handler));
                    }
                }
                catch (Exception ignore)
                {
                }
            }
            connectionListener.Stop();
            lock (this)
            {
                foreach (Session s in pending)
                {
                    s.Cancel();
                }
            }
            Log.FtpOffline(this);
        }

        public class DataStreamHandler
        {
            private enum ConnecitionType { NONE, PASV, PORT, EPSV, EPRT };

            private ConnecitionType connectionType = ConnecitionType.NONE;
            private TcpListener passiveListener = null;
            private String portHost = null;
            private int portPortnum = 0;
            private TcpClient activeClient = null;
            private System.Net.IPAddress addrIPv4 = null;
            private System.Net.IPAddress addrIPv6 = null;

            public DataStreamHandler()
            {
                String hostName = Dns.GetHostName();
                IPAddress[] myIPs = Dns.GetHostEntry(hostName).AddressList;
                foreach (IPAddress ip in myIPs)
                {
                    if (addrIPv4 == null && ip.AddressFamily == AddressFamily.InterNetwork)
                        addrIPv4 = ip;
                    if (addrIPv6 == null && ip.AddressFamily == AddressFamily.InterNetworkV6)
                        addrIPv6 = ip;
                }
            }

            public String SetPASV()
            {
                CloseAnyClient();
                if (passiveListener == null || passiveListener.LocalEndpoint.AddressFamily == AddressFamily.InterNetworkV6)
                    passiveListener = new TcpListener(IPAddress.Any, 0);
                passiveListener.Stop(); 
                passiveListener.Start(1);
                System.Net.IPEndPoint ipEndpoint = (System.Net.IPEndPoint)passiveListener.LocalEndpoint;
                int port = ipEndpoint.Port;
                String ip = Util.IPToString(addrIPv4).Replace('.', ',');
                int p2 = port % 256;
                int p1 = (port - p2) / 256;
                connectionType = ConnecitionType.PASV;
                String result = "(" + ip + "," + p1 + "," + p2 + ")";
                return result;
            }

            public String SetEPSV(String arg, AddressFamily controlProtocol)
            {
                if (arg == "1" || (arg == null && controlProtocol == AddressFamily.InterNetwork))
                {
                    SetPASV();
                    System.Net.IPEndPoint ipEndpoint = (System.Net.IPEndPoint)passiveListener.LocalEndpoint;
                    int port = ipEndpoint.Port;
                    connectionType = ConnecitionType.EPSV;
                    if (arg == null)
                        return "(|||" + port + "|)";
                    return "(|1|" + Util.IPToString(addrIPv4) + "|" + port + "|)";
                }
                else
                {
                    CloseAnyClient();
                    if (passiveListener == null || passiveListener.LocalEndpoint.AddressFamily == AddressFamily.InterNetwork)
                    {
                        //passiveListener = new TcpListener(addrIPv6, 0);
                        passiveListener = new TcpListener(IPAddress.IPv6Any, 0);
                        passiveListener.Stop();
                        passiveListener.Start(1);
                    }
                    System.Net.IPEndPoint ipEndpoint = (System.Net.IPEndPoint)passiveListener.LocalEndpoint;
                    int port = ipEndpoint.Port;

                    connectionType = ConnecitionType.EPSV;
                    if (arg == null || arg.Trim() == "")
                        return "(|||" + port + "|)";
                    else
                        return "(|2|" + Util.IPToString(addrIPv6) + "|" + port + "|)";
                }
            }

            public void SetPORT(String portStr)
            {
                CloseAnyClient();
                try
                {
                    if (portStr.IndexOf('(') >= 0 && portStr.IndexOf(')') >= 0)
                    {
                        portStr = portStr.Substring(portStr.IndexOf('(') + 1);
                        portStr = portStr.Substring(0, portStr.IndexOf(')'));
                    }
                    String[] portInfo = portStr.Split(",".ToCharArray());
                    byte h1 = byte.Parse(portInfo[0]);
                    byte h2 = byte.Parse(portInfo[1]);
                    byte h3 = byte.Parse(portInfo[2]);
                    byte h4 = byte.Parse(portInfo[3]);
                    int p1 = int.Parse(portInfo[4]);
                    int p2 = int.Parse(portInfo[5]);
                    portHost = "" + h1 + "." + h2 + "." + h3 + "." + h4;
                    portPortnum = 256 * p1 + p2;
                    connectionType = ConnecitionType.PORT;
                }
                catch (Exception)
                {
                    throw new FtpException(501, "Syntax error in parameters or arguments.");
                }
            }

            public void SetEPRT(String portStr)
            {
                try
                {
                    String[] args = portStr.Split("|".ToCharArray());
                    if (args[1] == "1")
                    {
                        String ip = args[2].Replace('.', ',');
                        int port = int.Parse(args[3]);
                        int p2 = port % 256;
                        int p1 = (port - p2) / 256;
                        SetPORT("(" + ip + "," + p1 + "," + p2 + ")");
                    }
                    else
                    {
                        CloseAnyClient();
                        portHost = args[2];
                        portPortnum = int.Parse(args[3]);
                    }
                    connectionType = ConnecitionType.EPRT;
                }
                catch (Exception)
                {
                    throw new FtpException(501, "Syntax error in parameters or arguments.");
                }
            }

            internal void CloseAnyClient()
            {
                if (connectionType == ConnecitionType.PASV || connectionType == ConnecitionType.PORT)
                    connectionType = ConnecitionType.NONE;
                if (activeClient != null && activeClient.Connected)
                    activeClient.Close();
            }

            internal void StopAnyListener()
            {
                try
                {
                    if (passiveListener != null)
                        passiveListener.Stop();
                }
                catch (Exception ignore) { }
            }

            public Stream GetDataStream()
            {
                //TBD: Handle when client does not connect - i.e. timeout and stuff
                if (connectionType == ConnecitionType.PASV)
                {
                    CloseAnyClient();
                    activeClient = passiveListener.AcceptTcpClient();
                    passiveListener.Stop();
                    return activeClient.GetStream();
                }
                else if (connectionType == ConnecitionType.PORT)
                {
                    CloseAnyClient();
                    activeClient = new TcpClient(portHost, portPortnum);
                    return activeClient.GetStream();
                }
                else if (connectionType == ConnecitionType.EPSV)
                {
                    CloseAnyClient();
                    activeClient = passiveListener.AcceptTcpClient();
                    //passiveListener.Stop();
                    return activeClient.GetStream();
                }
                else if (connectionType == ConnecitionType.EPRT)
                {
                    CloseAnyClient();
                    activeClient = new TcpClient(portHost, portPortnum);
                    return activeClient.GetStream();
                }
                else
                {
                    throw new FtpException(425, "Use PORT or PASV first.");
                }
            }
        }

        public class FtpException : Exception
        {
            public FtpException(int errNum, String msg)
                : base("" + errNum + " " + msg)
            {
            }
        }
        
        private class Session
        {
            private FtpServer server;
            private TcpClient tcpClient;
            private FtpHandler handler;
            private DataStreamHandler dataStreamHandler;
            private System.Net.IPAddress clientIP;
            private volatile bool running = false;


            public Session(FtpServer argServer, TcpClient argTcpClient, FtpHandler argHandler)
            {
                server = argServer;
                tcpClient = argTcpClient;
                handler = argHandler;
                clientIP = ((System.Net.IPEndPoint)tcpClient.Client.RemoteEndPoint).Address;
                dataStreamHandler = new DataStreamHandler();
                ThreadPool.QueueUserWorkItem(new WaitCallback(this.RunThread));
            }



            private void RunThread(Object stateInfo)
            {
                lock (this)
                {
                    if (handler == null)
                    {
                        return;
                    }
                    running = true;
                }
                NetworkStream netStream = null;
                StreamWriter commandOut = null;
                String user = null;
                String pass = null;

                try
                {
                    netStream = tcpClient.GetStream();
                    commandOut = Util.GetStreamWriter(netStream);
                    commandOut.AutoFlush = true;
                    commandOut.Write("220 " + Util.toolNameLink + "\r\n");
                    while (true)
                    {
                        String s = Util.ReadLineFromStream(netStream).Trim();
                        String sLow = s.ToLower();
                        if (sLow.StartsWith("user "))
                        {
                            user = s.Substring(s.IndexOf(' ') + 1);
                            commandOut.Write("331 Please specify the password.\r\n");
                        }
                        else if (sLow.StartsWith("pass "))
                        {
                            if (user != null)
                            {
                                pass = s.Substring(s.IndexOf(' ') + 1);
                                commandOut.Write("230 Login successful.\r\n");
                            }
                            else
                            {
                                commandOut.Write("503 Login with USER first.\r\n");
                            }
                        }
                        else if (sLow.StartsWith("quit"))
                        {
                            commandOut.Write("221 Goodbye.\r\n");
                            break;
                        }
                        else
                        {
                            commandOut.Write("530 Please login with USER and PASS.\r\n");
                        }
                        if (user != null && pass != null)
                            break;
                    }
                    if (user != null && pass != null)
                    {
                        Log.FtpConnected(clientIP, user, pass);
                        handler.Handle(user, pass, netStream, dataStreamHandler, clientIP);
                    }
                    return;
                }
                catch (Exception ex)
                {
                    try {
                        commandOut.Write("421 An error occured, closing control connection.\r\n"); //try notify client that we are closing things down
                    } catch (Exception) {}
                }
                finally
                {
                    server.Closed(this);
                    tcpClient.Close();
                    dataStreamHandler.CloseAnyClient();
                    dataStreamHandler.StopAnyListener();
                    Log.FtpClosed(clientIP, user, pass);
                }
            }


            public void Cancel()
            {
                try
                {
                    lock (this)
                    {
                        if (running == false) //cancel the thread before it even begins executing code body
                        {
                            handler = null; //null signals RunThread-method to return immediately after it is invoked
                            tcpClient.Close(); //close connection - we dont care
                        }
                        else
                        {
                            //close connection and let handler deal with any errors that might follow
                            tcpClient.Close();
                            dataStreamHandler.CloseAnyClient();
                        }
                    }
                }
                catch (Exception ignore)
                {
                }
            }

        }
    }
}
