﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net.Sockets;
using System.Net;

namespace FileServant
{
    public class HttpFileHandler : HttpServer.HttpHandler
    {
        private FileStore fileStore;
        private IPAddress clientIP;

        public HttpFileHandler(FileStore argFileStore)
        {
            fileStore = argFileStore;
        }

        private void Serve404(Stream responseStream, String url)
        {
            String response = "<html><head><title>404 Not Found</title></head><body><h1>Not Found</h1><p>The requested URL " + url + " was not found on this server.</p></body></html>";
            byte[] responseBytes = Util.StringToUTF8Bytes(response);
            List<String> headers = new List<String>();
            headers.Add("HTTP/1.1 404 Not Found");
            headers.Add("Content-Length: " + responseBytes.Length);
            headers.Add("Connection: close");
            HttpServer.WriteResponseHeaders(headers, responseStream);
            StreamWriter writer = Util.GetStreamWriter(responseStream);
            writer.Write(response);
            writer.Flush();
        }

        private void Serve400(Stream responseStream, String message)
        {
            String response = "<html><head><title>400 Bad Request</title></head><body><h1>Bad Request</h1><p>" + message + "</p></body></html>";
            byte[] responseBytes = Util.StringToUTF8Bytes(response);
            List<String> headers = new List<String>();
            headers.Add("HTTP/1.1 400 Bad Request");
            headers.Add("Content-Length: " + responseBytes.Length);
            headers.Add("Connection: close");
            HttpServer.WriteResponseHeaders(headers, responseStream);
            StreamWriter writer = Util.GetStreamWriter(responseStream);
            writer.Write(response);
            writer.Flush();
        }


        public void Handle(HttpServer.RequestMethod requestMethod, String url, List<String> requestHeaders, Stream requestBody, Stream responseStream, IPAddress argClientIP)
        {
            FileStore.URL fileStoreURL = new FileStore.URL(url);
            clientIP = argClientIP;
            String authUser = null;
            String authPass = null;
            foreach (String s in requestHeaders)
            {
                if (s.StartsWith("Authorization:"))
                {
                    byte[] bytes = Convert.FromBase64String(s.Substring(s.LastIndexOf(' ') + 1));
                    char[] chars = new char[bytes.Length];
                    for (int i = 0; i < bytes.Length; i++)
                        chars[i] = Convert.ToChar(bytes[i]);
                    String decodedString = new String(chars);
                    int separatorIndex = decodedString.IndexOf(':');
                    if (separatorIndex < 0)
                        break;
                    authUser = decodedString.Substring(0, separatorIndex);
                    authPass = decodedString.Substring(separatorIndex + 1);
                }
            }

            Log.HttpRequest(clientIP, url, requestMethod, authUser, authPass);

            if (requestMethod == HttpServer.RequestMethod.HEAD || requestMethod == HttpServer.RequestMethod.GET)
            {
                FileStore.VirtualFile vf = null;
                bool authRequired = false;
                try
                {
                    vf = fileStore.GetFile(fileStoreURL, authUser, authPass, false);
                }
                catch (Exception ex)
                {
                    authRequired = ex is FileStore.AuthenticationRequired;
                }

                if (url.ToLower() == "/favicon.ico")
                {
                    vf = new FileStore.VirtualFile("/favicon.ico", FileStore.VirtualFile.filetype.file, null, null);
                }

                if (authRequired)
                { //request authentication info from user
                    String body = "<html><head><title>" + WebUtility.HtmlEncode(fileStoreURL.GetURL()) + "</title><body><pre>\n";
                    body += Util.toolNameHtmlLink + "\n\n";
                    body += "Blocked: <a href=\"/\">/</a> " + WebUtility.HtmlEncode(fileStoreURL.GetURL().Substring(1)) + "\n\n";
                    body += "Authentication required. Provide correct username and password to access resource!\n\n";
                    body += "</pre></body></html>";

                    byte[] bodyBytes = Util.StringToUTF8Bytes(body);
                    List<String> headers = new List<String>();

                    headers.Add("HTTP/1.1 401 Authorization Required");
                    headers.Add("WWW-Authenticate: Basic realm=\"" + Util.toolName + " - protected resource\"");
                    headers.Add("Content-Type: text/html; charset=utf-8");
                    headers.Add("Content-Length: " + bodyBytes.Length);
                    headers.Add("Connection: close");
                    HttpServer.WriteResponseHeaders(headers, responseStream);
                    if (requestMethod != HttpServer.RequestMethod.HEAD)
                    {
                        StreamWriter bodyWriter = Util.GetStreamWriter(responseStream);
                        bodyWriter.Write(body);
                        bodyWriter.Flush();
                    }
                }
                else if (vf == null) //notify user that file cannot be found
                    Serve404(responseStream, url);
                else if (vf.type == FileStore.VirtualFile.filetype.directory)
                {
                    Log.DirServe(clientIP, vf);

                    String myUrl = fileStoreURL.GetURL();
                    //serve directory listing
                    List<FileStore.VirtualFile> dirContents = fileStore.ListFiles(fileStoreURL, authUser, authPass);
                    String body = "<html><head><title>" + WebUtility.HtmlEncode(myUrl) + "</title><body><pre>\n";
                    body += Util.toolNameHtmlLink + "\n\n";
                    body += "Serving: <a href=\"/\">/</a> ";
                    for (int i = 1, j = 1; i >= 0 && i < myUrl.Length; )
                    {
                        j = myUrl.IndexOf('/', i + 1);
                        if (j > 0)
                            body += "<a href=\"" +  Util.UrlEncode(myUrl.Substring(0, j)) + "\">" + WebUtility.HtmlEncode(myUrl.Substring(i, j - i)) + "</a> ";
                        else
                            body += "<a href=\"" + Util.UrlEncode(myUrl) + "\">" + WebUtility.HtmlEncode(myUrl.Substring(i)) + "</a>";
                        i = j;
                    }
                    body += "\n\n";
                    if (vf.permission == FileStore.Permission.readwrite)
                    {
                        body += "<form action=\"" + vf.virtualPath + "\" method=\"post\" enctype=\"multipart/form-data\"><input type=\"file\" name=\"file\"/><input type=\"submit\" value=\"Upload File\"/></form>";
                        body += "<form action=\"/\" method=\"post\"><input type=\"hidden\" name=\"action\" value=\"createdir\"/><input type=\"hidden\" name=\"parameter1\" value=\"" + Util.UrlEncode(vf.virtualPath) + "\"/><input type=\"text\" name=\"parameter2\" value=\"\"/><input type=\"submit\" value=\"Create Directory\" onclick=\"return confirm('Really create new directory?')\" /></form></td>";
                    }

                    int divID = 0;
                    foreach (FileStore.VirtualFile f in dirContents)
                    {
                        String date = f.GetTime().ToShortDateString() + " " + f.GetTime().ToShortTimeString();
                        String size = f.type == FileStore.VirtualFile.filetype.file ? "" + f.GetSize() : "<dir>";
                        String name = f.name;
                        size = WebUtility.HtmlEncode(size.PadLeft(15, ' '));
                        divID++;
                        if (f.permission == FileStore.Permission.readwrite)
                            body += "<input type=\"button\" value=\"&gt;\" onclick=\"document.getElementById('d" + divID + "').style.display = ''\"/>";

                        body += date + "   " + size + " <a href=\"" + Util.UrlEncode(f.virtualPath) + "\">" + WebUtility.HtmlEncode(name) + "</a>";

                        if (f.permission == FileStore.Permission.readwrite) {
                            body += "<div style=\"display: none\" id=\"d" + divID + "\"><table><tr>";

                            body += "<td><form><input type=\"button\" value=\"&lt;\" onclick=\"document.getElementById('d" + divID + "').style.display = 'none'\"/>&nbsp;&nbsp;</form></td>";

                            //rename button
                            body += "<td><form action=\"/\" method=\"post\"><input type=\"hidden\" name=\"action\" value=\"rename\"/><input type=\"hidden\" name=\"parameter1\" value=\"" + Util.UrlEncode(f.virtualPath) + "\"/><input type=\"test\" name=\"parameter2\" value=\"" + f.name + "\"/><input type=\"submit\" value=\"Rename\" onclick=\"return confirm('Really rename " + f.name + "?')\" /></form></td>";

                            //delete button
                            body += "<td><form action=\"/\" method=\"post\"><input type=\"hidden\" name=\"action\" value=\"delete\"/><input type=\"hidden\" name=\"parameter1\" value=\"" + Util.UrlEncode(f.virtualPath) + "\"/><input type=\"submit\" value=\"Delete\" onclick=\"return confirm('Really DELETE " + f.name + "?')\"/></form></td>";


                            body += "</tr></table></div>";
                        }
                        body += "\n";

                    }
                    body += "</pre></body></html>";
                    byte[] bodyBytes = Util.StringToUTF8Bytes(body);
                    List<String> headers = new List<String>();
                    headers.Add("HTTP/1.1 200 OK");

                    //these headers are to encourage browser to not cache information
                    headers.Add("Expires: Tue, 03 Jul 2001 06:00:00 GMT");
                    headers.Add("Cache-Control: no-store, no-cache, must-revalidate, max-age=0");
                    headers.Add("Cache-Control: post-check=0, pre-check=0");
                    headers.Add("Pragma: no-cache");

                    headers.Add("Content-Type: text/html; charset=utf-8");
                    headers.Add("Content-Length: " + bodyBytes.Length);
                    headers.Add("Connection: close");
                    HttpServer.WriteResponseHeaders(headers, responseStream);

                    if (requestMethod != HttpServer.RequestMethod.HEAD)
                    {
                        StreamWriter bodyWriter = Util.GetStreamWriter(responseStream);
                        bodyWriter.Write(body);
                        bodyWriter.Flush();
                    }
                }
                else
                {
                    //serve raw file contents

                    long size = 0;
                    BufferedStream readStream = null;
                    if (vf.virtualPath == "/favicon.ico")
                    {
                        System.Drawing.Icon favicon = Properties.Resources.icon;
                        MemoryStream ms = new MemoryStream();
                        favicon.Save(ms);
                        size = ms.Length;
                        ms.Position = 0;
                        readStream = new BufferedStream(ms);
                    }
                    else
                    {
                        Log.FileServe(clientIP, vf);
                        size = vf.GetSize();
                        readStream = new BufferedStream(fileStore.ReadFile(vf, authUser, authPass));
                    }
                    try
                    {
                        //determin range request
                        bool rangeRequest = false;
                        long byteFrom = 0; //(we note ourselves that byte indices are inclusive!)
                        long byteTo = size - 1;
                        long byteTotal = size;
                        try
                        {
                            foreach (String s in requestHeaders)
                            {
                                if (s.StartsWith("Range: bytes="))
                                {
                                    String bytePart = s.Substring(s.IndexOf('=') + 1);
                                    if (bytePart.IndexOf(',') >= 0) //we dont support multiple regions
                                        break;
                                    if (bytePart.EndsWith("-"))
                                    {
                                        byteFrom = long.Parse(bytePart.Substring(0, bytePart.Length - 1));
                                    }
                                    else
                                    {
                                        byteFrom = long.Parse(bytePart.Substring(0, bytePart.IndexOf('-')));
                                        long tmpByteTo = long.Parse(bytePart.Substring(bytePart.IndexOf('-') + 1));
                                        if (tmpByteTo < byteTo)
                                            byteTo = tmpByteTo; //we cant read past end of file
                                    }
                                    rangeRequest = (byteFrom <= byteTo); //only serve syntactically sane requests
                                    byteTotal = byteTo - byteFrom + 1;
                                    break;
                                }
                            }
                        }
                        catch (Exception) //ignore range requests if any errors occur when parsing range header
                        {
                            rangeRequest = false;
                        }

                        List<String> headers = new List<String>();
                        if (rangeRequest)
                        {
                            headers.Add("HTTP/1.1 206 Partial Content");
                            headers.Add("Content-Length: " + byteTotal);
                            headers.Add("Content-Range: bytes " + byteFrom + "-" + byteTo + "/" + size);
                        }
                        else
                        {
                            headers.Add("HTTP/1.1 200 OK");
                            headers.Add("Content-Length: " + size);
                        }

                        //give browsers a chance to detect and take propper actions if file has been modified in between a resuming download
                        headers.Add("Last-Modified: " + vf.GetTime().ToString("r"));
                        

                        headers.Add("Accept-Ranges: bytes");
                        headers.Add("Content-Type: application/octet-stream");
                        headers.Add("Connection: close");
                        HttpServer.WriteResponseHeaders(headers, responseStream);

                        if (requestMethod != HttpServer.RequestMethod.HEAD)
                        {
                            if (rangeRequest) //fast-forward to required bytes
                                readStream.Seek(byteFrom, SeekOrigin.Begin);
                            else
                                byteTotal = size;
                            BufferedStream bodyStream = new BufferedStream(responseStream);
                            byte[] buffer = new byte[524288]; //new byte[4096];
                            long bytes = 0;
                            while (bytes < byteTotal && (vf.share == null || vf.share.online))
                            {
                                int read = readStream.Read(buffer, 0, buffer.Length);
                                if (read == 0)
                                    break; //break out of loop - end of file reached - if this was before byteTotal we dont care and fail silently
                                bodyStream.Write(buffer, 0, read);
                                bytes += read;
                            }
                            bodyStream.Flush();
                        }
                    }
                    finally
                    {
                        readStream.Close();
                    }
                }
                return;
            }
            else if (requestMethod == HttpServer.RequestMethod.POST)
            {
                long contentLength = 0;
                bool formData = false;
                bool urlEncoded = false;
                String boundary = null;
                String contentTransferEncoding = "binary"; //7bit seems to be default encoding when no Content-Transfer-Encoding header is mentioned
                foreach (String s in requestHeaders) {
                    if (s.StartsWith("Content-Length: "))
                        contentLength = long.Parse(s.Substring(s.IndexOf(' ') + 1));
                    if (s.StartsWith("Content-Type: application/x-www-form-urlencoded"))
                        urlEncoded = true;
                    if (s.StartsWith("Content-Transfer-Encoding: "))
                    {
                        contentTransferEncoding = s.Substring(s.IndexOf(' ') + 1).Trim().ToLower();
                    }
                    if (s.StartsWith("Content-Type: multipart/form-data; boundary"))
                    {
                        formData = true;
                        boundary = s.Substring(s.IndexOf('=') + 1);
                    }
                }
                String redirect = null;
                bool success = false;
                if (url == "/" && urlEncoded)
                {
                    String command = "";
                    long read = 0;
                    //requestBody.ReadTimeout = 1000;
                    while (read < contentLength && read < 4096)
                    {
                        read++;
                        int readByte = requestBody.ReadByte();
                        if (readByte < 0)
                            break;
                        command += (char)readByte;
                    }
                    String action = null;
                    String parameter1 = null;
                    String parameter2 = null;

                    foreach (String s in command.Split("&".ToCharArray()))
                    {
                        if (s.StartsWith("action="))
                        {
                            action = Uri.UnescapeDataString(s.Substring(s.IndexOf('=') + 1));
                        }
                        else if (s.StartsWith("parameter1="))
                        {
                            //parameters are double-encoded as browser encodes space characters into '+' characters which is an error
                            parameter1 = Uri.UnescapeDataString(Uri.UnescapeDataString(s.Substring(s.IndexOf('=') + 1))); 
                        }
                        else if (s.StartsWith("parameter2="))
                        {
                            parameter2 = Uri.UnescapeDataString(s.Substring(s.IndexOf('=') + 1));
                        }
                    }
                    if (action == "createdir" && parameter1 != null && parameter2 != null)
                    {
                        parameter2 = parameter2.Trim();
                        if (parameter2 == "")
                        {
                            Serve400(responseStream, "No directory name specified.");
                            return;
                        }

                        FileStore.URL parentURL = new FileStore.URL(parameter1);
                        if (parentURL.GetURL() == "/")
                        {
                            Serve400(responseStream, "Cannot create root elements.");
                            return;
                        }

                        FileStore.VirtualFile parentDir = fileStore.GetFile(parentURL, authUser, authPass, false);
                        if (parentDir.permission == FileStore.Permission.read)
                        {
                            Serve400(responseStream, "Error: read only " + parentDir.virtualPath);
                            return;
                        }
                        FileStore.URL targetURL = new FileStore.URL(parentURL.GetURL() + "/" + parameter2.Replace("/", "").Replace("\\", "").Replace(":", ""));

                        try
                        {
                            FileStore.VirtualFile vf = fileStore.CreateDirectory(targetURL, authUser, authPass);
                            Log.FileWrite(clientIP, vf);
                            redirect = parentDir.virtualPath;
                            success = true;
                        }
                        catch (Exception)
                        {
                            Serve400(responseStream, "Could not create directory " + targetURL.GetURL());
                            return;
                        }
                    }
                    else if (action == "rename" && parameter1 != null && parameter2 != null)
                    {
                        parameter2 = parameter2.Replace('+', ' ');

                        parameter2 = parameter2.Trim();
                        if (parameter2 == "")
                        {
                            Serve400(responseStream, "No filename specified.");
                            return;
                        }


                        FileStore.URL fromURL = new FileStore.URL(parameter1);
                        FileStore.URL parentURL = new FileStore.URL(fromURL.GetURL().Substring(0, fromURL.GetURL().LastIndexOf('/')));
                        FileStore.URL toURL = new FileStore.URL(parentURL.GetURL() + "/" + parameter2.Replace("/", "").Replace("\\", "").Replace(":", ""));
                        if (parentURL.GetURL() == "/")
                        {
                            Serve400(responseStream, "Cannot rename root elements.");
                            return;
                        }
                        try
                        {
                            FileStore.VirtualFile fromFile = fileStore.GetFile(fromURL, authUser, authPass, false);
                            if (fromFile.permission == FileStore.Permission.read)
                            {
                                Serve400(responseStream, "Error: read only " + fromFile.virtualPath);
                                return;
                            }
                            FileStore.VirtualFile toParent = fileStore.GetFile(parentURL, authUser, authPass, false); //check parent directory for target exists
                            FileStore.VirtualFile toFile = null;
                            try
                            {
                                toFile = fileStore.GetFile(toURL, authUser, authPass, false); //check permissions and existence of target
                            }
                            catch (FileStore.VirtualPathDoesNotExist) { }
                            if (toFile != null) //error - target already exists
                            {
                                Serve400(responseStream, "Target already exists " + toFile.virtualPath);
                                return;
                            }
                            toFile = fileStore.GetFile(toURL, authUser, authPass, true);
                            //all checks passed - perform move
                            fileStore.Rename(fromURL, toURL, authUser, authPass);
                            Log.FileWrite(clientIP, fromFile);
                            Log.FileWrite(clientIP, toFile);
                            redirect = fromURL.GetURL().Substring(0, fromURL.GetURL().LastIndexOf('/'));
                            success = true;
                        }
                        catch (Exception ex)
                        {
                            Serve400(responseStream, "Move failed.");
                            return;
                        }
                    }
                    else if (action == "delete" && parameter1 != null)
                    {
                        FileStore.URL delURL = new FileStore.URL(parameter1);
                        FileStore.VirtualFile vf = fileStore.GetFile(delURL, authUser, authPass, false);
                        if (vf.permission == FileStore.Permission.read || vf.virtualPath.LastIndexOf('/') == 0)
                        {
                            Serve400(responseStream, "Error: read only " + vf.virtualPath);
                            return;
                        }
                        try
                        {
                            Log.FileWrite(clientIP, vf);
                            fileStore.Delete(vf, authUser, authPass);
                            redirect = delURL.GetURL().Substring(0, delURL.GetURL().LastIndexOf('/'));
                            success = true;
                        }
                        catch (Exception) { }
                    }
                    if (!success)
                    {
                        Serve400(responseStream, "Operation failed.");
                        return;
                    }
                }
                else if (formData)
                {
                    //we are uploading a file!
                    boundary = "\r\n--" + boundary;
                    long bodyRead = 0;
                    
                    //consume body until two times carriage return and newline appears
                    String tmp = "";
                    while (bodyRead < contentLength)
                    {
                        int readByte = requestBody.ReadByte(); bodyRead++;
                        tmp += (char)readByte;
                        if (tmp.EndsWith("\r\n\r\n"))
                            break;
                    }
                    //extract information about Content-Transfer-Encoding for data part
                    if (tmp.IndexOf("Content-Transfer-Encoding: ") > 0)
                    {
                        int cteIndex = tmp.IndexOf("Content-Transfer-Encoding: ");
                        contentTransferEncoding = tmp.Substring(cteIndex + 27);
                        contentTransferEncoding = contentTransferEncoding.Substring(0, contentTransferEncoding.IndexOf('\r')).ToLower().Trim();
                    }
                    //extract filename from consumed bytes
                    if (tmp.IndexOf(" filename=\"") < 0)
                    {
                        Serve400(responseStream, "Bad Request.");
                        return;
                    }
                    String fileName = "";
                    if (tmp.IndexOf(" filename=\"") >= 0)
                    {
                        fileName = tmp.Substring(tmp.IndexOf(" filename=\"") + 11);
                        fileName = fileName.Substring(0, fileName.IndexOf('"')).Replace("/", "").Replace("\\", "").Replace(":", "").Trim();
                    }
                    if (fileName == "")
                    {
                        Serve400(responseStream, "No file submitted.");
                        return;
                    }

                    if (url.Length <= 1)
                    {
                        Serve400(responseStream, "Cannot upload to root directory");
                        return;
                    }
                    FileStore.VirtualFile target = fileStore.GetFile(new FileStore.URL(url + "/" + fileName), authUser, authPass, true);
                    if (target.permission == FileStore.Permission.read)
                    {
                        Serve400(responseStream, "Error: read only " + target.virtualPath);
                        return;
                    }
                    

                    /*
                    "7bit" / "8bit" / "binary" /
                  "quoted-printable" / "base64"

                     */                    

                    System.Text.UTF8Encoding encoder = new UTF8Encoding();
                    byte[] boundaryBytes = encoder.GetBytes(boundary);

                    //some effort to prevent user from corrupting files because of encoding schemes unknown to this implementation
                    if (contentTransferEncoding != "binary")
                    {
                        Serve400(responseStream, "Error does not support Content-Transfer-Encoding of \"" + contentTransferEncoding + "\". Only \"binary\" is supported.");
                        return;
                    }
                    
                    Stream fs = fileStore.WriteFile(target, 0, authUser, authPass);
                    Log.FileWrite(clientIP, target);
                    BufferedStream targetStream = new BufferedStream(fs);
                    byte[] bufferedBytes = new byte[524288]; //new byte[4096];
                    int numRead = 0;
                    try
                    {
                        int boundMatch = 0;
                        while (bodyRead < contentLength && target.share.online)
                        {
                            numRead = requestBody.Read(bufferedBytes, 0, bufferedBytes.Length);
                            bodyRead += numRead;

                            //TBD: make detection of boundary less cpu intensive
                            for (int bPos = 0; bPos < numRead; bPos++)
                            {
                                byte readByte = bufferedBytes[bPos];
                                //Console.Write((char)readByte);
                                if (boundaryBytes[boundMatch] == readByte)
                                    boundMatch++;
                                else
                                {
                                    if (boundMatch != 0)
                                    {
                                        for (int i = 0; i < boundMatch; i++)
                                        {
                                            targetStream.WriteByte(boundaryBytes[i]);
                                        }
                                        boundMatch = 0;
                                        targetStream.WriteByte(readByte);
                                    }
                                    else
                                    {
                                        targetStream.WriteByte(readByte);
                                    }
                                }
                                if (boundMatch == boundaryBytes.Length)
                                {
                                    //consume rest of stream
                                    while (bodyRead < contentLength)
                                    {
                                        numRead = requestBody.Read(bufferedBytes, 0, bufferedBytes.Length);
                                        bodyRead += numRead;
                                    }
                                    break;
                                }
                            }
                        }
                        redirect = target.virtualPath.Substring(0, target.virtualPath.LastIndexOf('/'));
                        success = true;
                    }
                    catch (Exception ex)
                    {
                        return;
                    }
                    finally
                    {
                        targetStream.Flush();
                        targetStream.Close();
                    }
                    if (!success)
                    {
                        Serve400(responseStream, "Upload failed.");
                        return;
                    }
                }
                else
                {
                    Serve400(responseStream, "Bad Request.");
                    return;
                }
                //serve success!
                String body = "<html><head><title>Success!</title><body><pre>\n";
                body += "Operation successfull!\nGo to <a href=\"" + Util.UrlEncode(redirect) + "\">" + System.Net.WebUtility.HtmlEncode(redirect) + "</a>";
                body += "</pre></body></html>";
                byte[] bodyBytes = Util.StringToUTF8Bytes(body);
                List<String> headers = new List<String>();
                headers.Add("HTTP/1.1 303 See Other");
                //headers.Add("Refresh: 0; url=" + redirect);
                headers.Add("Location: " + redirect);

                //these headers are to encourage browser to not cache information
                headers.Add("Expires: Tue, 03 Jul 2001 06:00:00 GMT");
                headers.Add("Cache-Control: no-store, no-cache, must-revalidate, max-age=0");
                headers.Add("Cache-Control: post-check=0, pre-check=0");
                headers.Add("Pragma: no-cache");

                headers.Add("Content-Type: text/html; charset=utf-8");
                headers.Add("Content-Length: " + bodyBytes.Length);
                headers.Add("Connection: close");
                HttpServer.WriteResponseHeaders(headers, responseStream);

                StreamWriter bodyWriter = Util.GetStreamWriter(responseStream);
                bodyWriter.Write(body);
                bodyWriter.Flush();

                return;
            }
            Serve400(responseStream, "Request failed");
        }
    }
}
