﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using System.IO;
using System.Threading;
using System.Web;

namespace FileServant
{
    public class HttpServer
    {
        private TcpListener connectionListener;
        private HttpHandler handler;
        public enum RequestMethod {HEAD, GET, POST};
        private List<Request> pending = new List<Request>();
        private Thread serverThread;
        private bool running = true;
        public int port = 0;
        public AddressFamily addressFamily;

        public HttpServer(TcpListener argConnectionListener, HttpHandler argHandler)
        {
            connectionListener = argConnectionListener;
            handler = argHandler;
            if (handler == null)
                throw new NullReferenceException();
            port = ((System.Net.IPEndPoint)connectionListener.LocalEndpoint).Port;
            addressFamily = connectionListener.LocalEndpoint.AddressFamily;
            serverThread = new Thread(new ThreadStart(this.RunServer));
            serverThread.Start();
        }

        public void KillServer()
        {
            running = false;
            //stop accepting new connections and kill any helper threads
            try
            {
                connectionListener.Stop(); 
            }
            catch (Exception ignore)
            {
            }

            //wait until all pending requests has been cancelled before returning
            int numPending = 1;
            while (numPending > 0)
            {
                lock (this)
                {
                    numPending = pending.Count;
                }
                if (numPending == 0)
                    break;
                Thread.Sleep(100); 
            }
        }

        public interface HttpHandler
        {
            void Handle(RequestMethod requestMethod, String url, List<String> requestHeaders, Stream requestBody, Stream responseStream, System.Net.IPAddress clientIP);
        }

        private void HasBeenServed(Request request)
        {
            lock (this)
            {
                pending.Remove(request);
            }
        }

        private void RunServer()
        {
            Log.HttpOnline(this);
            while (running)
            {
                try
                {
                    TcpClient client = connectionListener.AcceptTcpClient();
                    lock (this)
                    {
                        pending.Add(new Request(this, client, handler));
                    }
                } catch (Exception ignore) {
                }
            }
            connectionListener.Stop();
            lock (this)
            {
                foreach (Request r in pending)
                {
                    r.Cancel();
                }
            }
            Log.HttpOffline(this);
        }

        //utility method to be used in custom implementation of HttpServer.HttpHandler interface
        public static void WriteResponseHeaders(List<String> headers, Stream responseStream) {
            StreamWriter writer = Util.GetStreamWriter(responseStream);
            foreach (String s in headers)
            {
                writer.Write(s);
                writer.Write("\r\n");
            }
            writer.Write("\r\n"); //an empty line denotes separation between header data and body data
            writer.Flush();
        }

        private class Request
        {
            private HttpServer server;
            private TcpClient tcpClient;
            private HttpHandler handler;
            private volatile bool running = false;


            public Request(HttpServer argServer, TcpClient argTcpClient, HttpHandler argHandler)
            {
                server = argServer;
                tcpClient = argTcpClient;
                handler = argHandler;
                ThreadPool.QueueUserWorkItem(new WaitCallback(this.RunThread));
            }



            private void RunThread(Object stateInfo)
            {
                lock (this)
                {
                    if (handler == null)
                    {
                        return;
                    }
                    running = true;
                }
                NetworkStream netStream = null; 
                try
                {
                    netStream = tcpClient.GetStream();
                    //consume all headers
                    List<String> headers = new List<String>();
                    while (true)
                    {
                        String s = Util.ReadLineFromStream(netStream).Replace("\r", "");
                        if (s == "") //empty line - end of header portion
                            break;
                        headers.Add(s);
                    }

                    if (headers.Count > 0)
                    {
                        //extract request method and url from first header
                        String firstHeader = headers[0];
                        RequestMethod method = RequestMethod.HEAD;
                        if (firstHeader.StartsWith("HEAD "))
                            method = RequestMethod.HEAD;
                        else if (firstHeader.StartsWith("GET "))
                            method = RequestMethod.GET;
                        else if (firstHeader.StartsWith("POST "))
                            method = RequestMethod.POST;
                        else
                            ServeErrorAndThrow(404, "Request method not supported: " + firstHeader, netStream, true);
                        String url = firstHeader.Substring(firstHeader.IndexOf(' ') + 1);

                        url = Uri.UnescapeDataString(url.Substring(0, url.IndexOf(' ')));
                        //invoke handler to respond to request
                        handler.Handle(method, url, headers, netStream, netStream, ((System.Net.IPEndPoint)tcpClient.Client.RemoteEndPoint).Address);
                        return;
                    }
                    //client sent no headers with request?
                    ServeErrorAndThrow(404, null, netStream, true);
                }
                catch (Exception ex)
                {
                    if (netStream != null) 
                        ServeErrorAndThrow(500, null, netStream, false);
                }
                finally
                {
                    server.HasBeenServed(this);
                    tcpClient.Close();
                }
            }

            private void ServeErrorAndThrow(int errorCode, String errMsg, Stream responseStream, bool throwException)
            {
                try
                {
                    if (errMsg == null)
                        errMsg = "An error occured. Sorry.";
                    errMsg = System.Net.WebUtility.HtmlEncode(errMsg);
                    String response = "<html><head><title>ERROR " + errorCode + "</title></head><body><h1>ERROR " + errorCode + "</h1><p>" + errMsg + "</p></body></html>";

                    List<String> headers = new List<String>();
                    headers.Add("HTTP/1.1 " + errorCode + " " + errMsg);
                    headers.Add("Content-Length: " + response.Length);
                    headers.Add("Connection: close");
                    HttpServer.WriteResponseHeaders(headers, responseStream);
                    StreamWriter writer = Util.GetStreamWriter(responseStream);
                    writer.Write(response);
                    writer.Flush();

                }
                catch (Exception ex)
                {
                }
                if (throwException)
                    throw new Exception(errorCode + " Error served");
            }


            public void Cancel()
            {
                try
                {
                    lock (this)
                    {
                        if (running == false) //cancel the thread before it even begins executing code body
                        {
                            handler = null; //null signals RunThread-method to return immediately after it is invoked
                            tcpClient.Close(); //close connection - we dont care
                        }
                        else
                        {
                            //close connection and let handler deal with any errors that might follow
                            tcpClient.Close();
                        }
                    }
                }
                catch (Exception ignore)
                {
                }
            }
            
        }
    }
}
