﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace FileServant
{
    public partial class ManageForm : Form
    {
        private Control control;
        private UpdateCheck updateCheck;

        public ManageForm(Control argControl, String instantSharePath) : base()
        {
            control = argControl;
            InitializeComponent();
            if (instantSharePath != null && instantSharePath.LastIndexOf('\\') >= 0)
            {
                instantSharePath = instantSharePath.Trim();
                txtPath.Text = instantSharePath;
            }
        }

        private long latestLogIndex = 0;
        private String logFileName = Log.LogFileName();
        private List<FileStore.RootShare> currentViewed = new List<FileStore.RootShare>();
        BindingSource logBindingSource = new BindingSource();
        private bool autoName = true;


        private void AddLogLine(String l)
        {
            bool scrollLast = false;
            if (dataLogView.SelectedRows.Count == 1)
            {
                scrollLast = dataLogView.SelectedRows[0].Index == dataLogView.Rows.Count - 1;
            }
            else
                scrollLast = true;
            if (scrollLast)
            {
                dataLogView.ClearSelection();
            }

            String line = l;
            try
            {
                String[] headers = line.Split(" ".ToCharArray());
                String eventType = headers[2];
                DateTime date = Convert.ToDateTime(headers[0], System.Globalization.CultureInfo.CurrentCulture);
                DateTime time = Convert.ToDateTime(headers[1].Substring(0, headers[1].IndexOf(',')), System.Globalization.CultureInfo.CurrentCulture);
                int ms = int.Parse(headers[1].Substring(headers[1].IndexOf(',')+1));
                DateTime dateTime = new DateTime(date.Year, date.Month, date.Day, time.Hour, time.Minute, time.Second, ms);
                line = line.Substring(line.IndexOf(' ') + 1);
                line = line.Substring(line.IndexOf(' ') + 1);
                line = line.Substring(line.IndexOf(' ') + 1);
                int row = dataLogView.Rows.Add(new Object[] { dateTime, eventType, line});
                Color c = Color.White;
                if (eventType == "SERVICE") {
                    c = Color.White;
                } else if (eventType == "FILESYSTEM") {
                    if (line.IndexOf("WRITE_FILE") >= 0)
                        c = Color.Gold;
                    else
                        c = Color.PaleGoldenrod;
                } else if (eventType == "HTTP") {
                    c = Color.LightCyan;
                } else if (eventType == "FTP") {
                    c = Color.PaleTurquoise;
                } else if (eventType == "SERVER") {
                    if (line.IndexOf("ONLINE") >= 0)
                        c = Color.PaleGreen;
                    else
                        c = Color.PaleVioletRed;
                } else if (eventType == "SHARE") {
                    c = Color.LightGray; ;
                }
                dataLogView.Rows[row].DefaultCellStyle.BackColor = c;

            }
            catch (Exception)
            {
                //TBD: add something to log
                dataLogView.Rows.Add(new Object[] { "?", "?", l});
            }

            if (scrollLast)
            {
                dataLogView.Rows[dataLogView.Rows.Count - 1].Selected = true;
                dataLogView.FirstDisplayedScrollingRowIndex = dataLogView.Rows.Count - 1;
            }
             
            
        }


        private void WatchLogFile()
        {
            lock (this)
            {
                try
                {
                    FileInfo logFile = new FileInfo(logFileName);
                    if (logFile.Exists)
                    {
                        long logLength = logFile.Length;
                        if (logLength > latestLogIndex)
                        {
                            bool skipFirst = false;
                            if (logLength-latestLogIndex > 1024*10)
                            {
                                //TBD: have to do something about adding a lot of rows to a datagridview as it very slow 
                                //     this is just a hack 
                                skipFirst = true;
                                latestLogIndex = logLength - 1024*10;
                                dataLogView.Rows.Clear();
                            }
                            System.IO.Stream logStream = logFile.Open(FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
                            logStream.Seek(latestLogIndex, System.IO.SeekOrigin.Begin);
                            System.IO.StreamReader logReader = new System.IO.StreamReader(logStream);

                            String s;
                            while ((s = logReader.ReadLine()) != null)
                            {
                                if (skipFirst)
                                    AddLogLine("<Skipped first " + logStream.Position + " bytes of log file>");
                                else
                                    AddLogLine(s);
                                skipFirst = false;
                            }
                            latestLogIndex = logStream.Position;
                            logStream.Close();
                        }
                        else if (logLength < latestLogIndex)
                        {
                            latestLogIndex = 0;
                            dataLogView.Rows.Clear();
                        }
                    }
                    else if (latestLogIndex > 0)
                    {
                        latestLogIndex = 0;
                        dataLogView.Rows.Clear();
                    }
                }
                catch (Exception ex)
                {
                    latestLogIndex = 0;
                }
                dataLogView.AutoResizeColumns();
            }
        }


        private String CreateShareNameFromPath(String path)
        {
            if (path != "")
            {
                if (!path.Contains(":\\"))
                    return "";
                String name = "";
                try
                {
                    DirectoryInfo di = new DirectoryInfo(txtPath.Text);
                    name = new FileStore.URL(di.Name).GetURL();
                }
                catch (Exception)
                {
                    name = "MyShare";
                }
                return name;
            }
            return "";
        }

        private void CreateShare()
        {
            if (txtPath.Text != "" && txtPath.Text.Contains(":\\"))
            {
                String name = txtName.Text;
                if (name == "")
                    name = CreateShareNameFromPath(txtPath.Text);

                FileStore.RootShare share = control.CreateShare(new FileStore.URL(name), txtPath.Text, radioButtonReadWrite.Checked ? FileStore.Permission.readwrite : FileStore.Permission.read, txtUser.Text != "" ? txtUser.Text : null, txtPass.Text != "" ? txtPass.Text : null);
                if (share == null)
                    MessageBox.Show("Failed creating share!");
                else
                {
                    txtName.Text = "";
                    txtUser.Text = "";
                    txtPass.Text = "";
                    txtPath.Text = "";
                    radioButtonRead.Checked = true;
                    radioButtonReadWrite.Checked = false;
                    autoName = true;
                }
                RefreshListShares();
            }
        }

        private void RefreshListShares()
        {
            bool refresh = false;
            List<FileStore.RootShare> shares = control.ListShares();
            if (currentViewed.Count != shares.Count)
                refresh = true;
            if (!refresh)
            {
                for (int i = 0; i < shares.Count; i++)
                {
                    if (shares[i].ToString() != currentViewed[i].ToString())
                    {
                        refresh = true;
                        break;
                    }
                }
            }
            if (refresh)
            {
                currentViewed.Clear();
                listShares.Rows.Clear();
                foreach (Object o in shares)
                {
                    String s = o.ToString();
                    currentViewed.Add((FileStore.RootShare)o);
                    String[] splitted = s.Split("|".ToCharArray());
                    String[] row = new String[6];
                    row[0] = splitted[0];
                    row[1] = "true";
                    row[2] = splitted[2] == "rw" ? "true" : "false";
                    row[3] = splitted[1];
                    row[4] = splitted[3];
                    row[5] = splitted[4];
                    
                    listShares.Rows.Add(row);
                }
                txtDetails.Text = "";
                listShares.ClearSelection();
                buttonRemoveAll.Enabled = currentViewed.Count > 0;
            }
        }

        private void ManageForm_Load(object sender, EventArgs e)
        {
            listShares.SelectionChanged += new EventHandler(listShares_SelectionChanged);
            txtName.KeyPress += new KeyPressEventHandler(txtName_KeyPress);
            txtPath.KeyPress += new KeyPressEventHandler(txtPath_KeyPress);
            txtUser.KeyPress += new KeyPressEventHandler(txtUser_KeyPress);
            txtPass.KeyPress += new KeyPressEventHandler(txtPass_KeyPress);
            radioButtonRead.KeyPress += new KeyPressEventHandler(radioButtonRead_KeyPress);
            radioButtonReadWrite.KeyPress += new KeyPressEventHandler(radioButtonReadWrite_KeyPress);
            RefreshListShares();
            updateCheck = new UpdateCheck(false);
            this.Text = Util.toolName;
            lblVersion.Text = updateCheck.GetCurrentVersion().ToString();
        }

        void radioButtonReadWrite_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == '\r')
                CreateShare();
        }

        void radioButtonRead_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == '\r')
                CreateShare();
        }

        void txtPass_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == '\r')
                CreateShare();
        }

        void txtUser_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == '\r')
                CreateShare();
        }

        void txtPath_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == '\r')
                CreateShare();
        }

        void txtName_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == '\r')
                CreateShare();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            CreateShare();
        }

        private void radioButtonRead_CheckedChanged(object sender, EventArgs e)
        {
            radioButtonReadWrite.Checked = !radioButtonRead.Checked;
        }

        private void radioButtonReadWrite_CheckedChanged(object sender, EventArgs e)
        {
            radioButtonRead.Checked = !radioButtonReadWrite.Checked;
        }

        private void button4_Click(object sender, EventArgs e)
        {
            DialogResult result = folderBrowserDialog1.ShowDialog();
            if (result == System.Windows.Forms.DialogResult.OK)
            {
                txtPath.Text = folderBrowserDialog1.SelectedPath;
                txtName.Focus();
            }
        }

        private void txtName_TextChanged(object sender, EventArgs e)
        {
            buttonCreate.Enabled = txtPath.Text.Length > 0;
            if (txtName.Text.Length > 0)
            {
                String newName = CreateShareNameFromPath(txtPath.Text);
                if (newName == txtName.Text)
                    autoName = true;
                else
                    autoName = false;
            }
            else
            {
                autoName = true;
                txtName.Text = CreateShareNameFromPath(txtPath.Text);
            }
        }

        private void txtUser_TextChanged(object sender, EventArgs e)
        {
            lblUser.Enabled = lblPass.Enabled = (txtUser.Text.Length > 0 || txtPass.Text.Length > 0);
        }

        private void txtPass_TextChanged(object sender, EventArgs e)
        {
            lblUser.Enabled = lblPass.Enabled = (txtUser.Text.Length > 0 || txtPass.Text.Length > 0);
        }

        private void txtPath_TextChanged(object sender, EventArgs e)
        {
            buttonCreate.Enabled = txtPath.Text.Length > 0;
            if (autoName)
            {
                txtName.Text = CreateShareNameFromPath(txtPath.Text);
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            if (listShares.SelectedRows.Count != 0)
            {
                int selectedIndex = listShares.SelectedRows[0].Index;
                FileStore.RootShare share = (FileStore.RootShare)currentViewed[selectedIndex];
                if (share != null && control.DeleteShare(new FileStore.URL(share.rootDirName)) != "SUCCESS")
                    MessageBox.Show("Failed removing share!");
                txtDetails.Text = "";
                RefreshListShares();
                if (autoName)
                {
                    txtName.Text = share.rootDirName;
                    txtUser.Text = share.user != null ? share.user : "";
                    txtPass.Text = share.pass != null ? share.pass : "";
                    txtPath.Text = share.directory.FullName;
                    radioButtonRead.Checked = share.permission == FileStore.Permission.read;
                    radioButtonReadWrite.Checked = share.permission == FileStore.Permission.readwrite;
                    autoName = true;
                }
            }
        }

        private void button3_Click(object sender, EventArgs e)
        {
            if (control.DeleteAll() != "SUCCESS")
                MessageBox.Show("Failed removing all shares!");
            txtDetails.Text = "";
            RefreshListShares();
        }

        private void refreshShareListTimer_Tick(object sender, EventArgs e)
        {
            RefreshListShares();
        }

        private void logWatchTimer_Tick(object sender, EventArgs e)
        {
            WatchLogFile();
        }

        private void button1_Click_1(object sender, EventArgs e)
        {
            try
            {
                FileInfo logFile = new FileInfo(logFileName);
                if (logFile.Exists)
                {
                    Stream s = logFile.Open(FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);
                    s.SetLength(0);
                    s.Flush();
                    s.Close();
                }
            }
            catch (Exception) { }
        }

        void listShares_SelectionChanged(object sender, EventArgs e)
        {
            if (listShares.SelectedRows.Count > 0)
            {
                FileStore.RootShare share = currentViewed[listShares.SelectedRows[0].Index];
                int[] ports = control.GetFtpHttpPorts();
                txtDetails.Text = Program.GetShareInfo(share, ports[0], ports[1]);
                buttonRemove.Enabled = true;
            } else
                buttonRemove.Enabled = false;
        }

        private void updateTimer_Tick(object sender, EventArgs e)
        {
            if (updateCheck.CheckIsDone())
            {
                updateTimer.Enabled = false;
                try
                {
                    if (updateCheck.OnlineVersionAvailable())
                    {
                        String caption = "Update Available";
                        String message = "A new version of File Servant is available online! (version " + updateCheck.GetOnlineVersion().ToString() + ")\n" +
                            updateCheck.GetOnlineVersionURL() + "\n" +
                            "\nDo you want to download it?\n" +
                            "\n(remember to run the downloaded file to install new version)";

                        DialogResult result = MessageBox.Show(message, caption, MessageBoxButtons.YesNo);
                        if (result == DialogResult.Yes)
                        {
                            System.Diagnostics.Process.Start(updateCheck.GetOnlineVersionURL());
                        }

                    }
                }
                catch (Exception) { }
            }
        }

        private void btnHomepage_Click(object sender, EventArgs e)
        {
            
            try
            {
                System.Diagnostics.Process.Start("http://www.fileservant.net/");
            }
            catch (Exception)
            {
                MessageBox.Show("Could not open the browser, sorry.\n\nVisit the homepage by going to:\nhttp://www.FileServant.net", "Could not open browser, sorry!", MessageBoxButtons.OK);
            }
        }

        private void linkLabel1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            MessageBox.Show("File Servant is Copyright (c) 2010, Martin Karlsen.\n" +
                "All rights reserved.\n" + 
                "\n" + 
                "Redistribution and use in source and binary forms, with or without modification, are " + 
                "permitted provided that the following conditions are met:\n" + 
                "\n" + 
                "1. Redistributions of source code must retain the above copyright notice, this list of " + 
                "conditions and the following disclaimer.\n" + 
                "\n" + 
                "2. Redistributions in binary form must reproduce the above copyright notice, this list " + 
                "of conditions and the following disclaimer in the documentation and/or other materials " + 
                "provided with the distribution.\n" + 
                "\n" + 
                "THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND ANY EXPRESS " + 
                "OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF " + 
                "MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL " + 
                "THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, " + 
                "SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF " + 
                "SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) " + 
                "HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR " + 
                "TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, " + 
                "EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.",
                "Here's the License");
        }

        private void button2_Click_1(object sender, EventArgs e)
        {
            try
            {
                System.Diagnostics.Process.Start(logFileName);
            }
            catch (Exception)
            {
                MessageBox.Show("Could not open the log file, sorry.\nYou can inspect the file manually at:\n" + logFileName);
            }

        }

    }
}
