﻿/*
File Servant is Copyright (c) 2010, Martin Karlsen. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR 
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Text;
using System.Net.Sockets;
using System.Net;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.IO;
using System.Threading;

namespace FileServant
{
    public class Program
    {
        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool AllocConsole();

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool FreeConsole();

        [DllImport("kernel32", SetLastError = true)]
        static extern bool AttachConsole(int dwProcessId);

        [DllImport("user32.dll")]
        static extern IntPtr GetForegroundWindow();

        [DllImport("user32.dll", SetLastError = true)]
        static extern uint GetWindowThreadProcessId(IntPtr hWnd, out int lpdwProcessId);

        private static bool attached = false;

        private static UpdateCheck update = null;

        private static bool FromCommandPrompt()
        {
            Process me = Process.GetCurrentProcess();
            Process parent = null;

            IntPtr ptr = GetForegroundWindow();
            int u;
            GetWindowThreadProcessId(ptr, out u);
            parent = Process.GetProcessById(u);

            //determin if we are started from a command prompt
            bool fromCommandPrompt = parent.ProcessName == "cmd";
            return fromCommandPrompt;
        }

        private static void Attatch() {
            Process me = Process.GetCurrentProcess();
            Process parent = null;

            /*
            try
            {
                //method one query performance counter - slooooow
                PerformanceCounter pc = new PerformanceCounter("Process", "Creating Process Id", me.ProcessName);
                parent = Process.GetProcessById((int)pc.RawValue);
                MessageBox.Show("" + parent.Id);

            } catch (Exception) {
            }
             */

            //use second method - less reliable
            IntPtr ptr = GetForegroundWindow();
            int u;
            GetWindowThreadProcessId(ptr, out u);
            parent = Process.GetProcessById(u);

            //determin if we are started from a command prompt
            bool fromCommandPrompt = parent.ProcessName == "cmd";

            if (fromCommandPrompt)
            {
                AttachConsole(parent.Id);
                attached = true;
            }
        }

        private static void Detatch()
        {
            if (attached)
            {
                FreeConsole();
                attached = false;
            }
        }

        private static void WriteLine(String s) {
            if (attached)
            {
                Console.WriteLine(s);
            }
        }

        private static void Write(String s)
        {
            if (attached)
            {
                Console.Write(s);
            }
        }


        private static void WriteHeader()
        {
            if (attached)
            {
                Console.Write('\r' + new String(' ', Console.WindowWidth - 1) + '\r'); //clear current line in console. Usually cmd has written "c:\sdfasdf>" at this point
                Console.WriteLine(Util.toolName);
                Console.WriteLine("Serve files over FTP and HTTP - at the snap of a finger!");

                Console.Write("(checking version...)\r");
                update = new UpdateCheck(true);
                
                if (update.CheckIsDone())
                {
                    if (update.OnlineVersionAvailable())
                    {
                        Console.WriteLine("Version " + update.GetCurrentVersion().ToString());
                        Console.WriteLine("A newer version (" + update.GetOnlineVersion().ToString() + ") is available online!");
                        Console.WriteLine("Download update here: " + update.GetOnlineVersionURL());
                    }
                    else
                    {
                        Console.WriteLine("Version " + update.GetCurrentVersion().ToString() + " (latest version)");
                    }
                }
                else
                {
                    Console.WriteLine("Version " + update.GetCurrentVersion().ToString() + " (checking online for newer version...)");
                }
                Console.WriteLine("");
            }
        }

        private static void WriteFooter()
        {
            WriteLine("");
            Write(Directory.GetCurrentDirectory() + ">"); //fake the overwritten command prompt-looking line like "c:\asasdfsdaf>"
        }


        private static void WriteHelp()
        {
            WriteLine("List of command line options:");
            WriteLine("");
            FileInfo f = new FileInfo(Application.ExecutablePath);
            WriteLine("  Display this help message:");
            WriteLine("    " + f.Name + " | " + f.Name + " /? | " + f.Name + " /help");
            WriteLine("");
            WriteLine("  Summon 'point-and-click' interface:");
            WriteLine("    " + f.Name + " /window");
            WriteLine("");
            WriteLine("  Summon 'point-and-click' interface for sharing specific folder:");
            WriteLine("    " + f.Name + " PATH");
            WriteLine("   where");
            WriteLine("    PATH is the filesystem path passed to the 'point-and-click' interface");
            WriteLine("");
            WriteLine("  List online shares:");
            WriteLine("    " + f.Name + " /list");
            WriteLine("");
            WriteLine("  Create a share:");
            WriteLine("    " + f.Name + " /create SHARENAME PATH PERMISSION [USER PASSWORD]");
            WriteLine("   where");
            WriteLine("    SHARENAME is name used to identify share");
            WriteLine("    PATH is the file system path to be shared");
            WriteLine("    PERMISSION is type of permission to grant, specify either r or rw");
            WriteLine("    [USER PASSWORD]: optional pair of user name and password");
            WriteLine("");
            WriteLine("   example: Create a read-only share named 'Docs' to share files from folder");
            WriteLine("            'C:\\Users\\user\\My Documents' and require authentication");
            WriteLine("            with username 'me' and password 'mypass'");
            WriteLine("    " + f.Name + " /create MyDocuments \"C:\\Users\\user\\My Documents\" r me mypass");
            WriteLine("");
            WriteLine("  Remove a share:");
            WriteLine("    " + f.Name + " /remove SHARENAME");
            WriteLine("  or");
            WriteLine("    " + f.Name + " /removeall");

        }



        public static String GetShareInfo(FileStore.RootShare share, int ftpPort, int httpPort)
        {
            String hostIPv4 = null;
            String hostIPv6 = null;
            String hostName = Dns.GetHostName();
            IPAddress[] myIPs = Dns.GetHostEntry(hostName).AddressList;
            foreach (IPAddress ip in myIPs)
            {
                if (hostIPv4 == null && ip.AddressFamily == AddressFamily.InterNetwork)
                    hostIPv4 = Util.IPToString(ip);
                if (hostIPv6 == null && ip.AddressFamily == AddressFamily.InterNetworkV6)
                    hostIPv6 = Util.IPToString(ip);
            }
            String result = "-> Share name: " + share.rootDirName;
            if (share.user != null || share.pass != null)
                result += " USER/PASSWORD=(" + (share.user != null ? share.user : "") + ":" + (share.pass != null ? share.pass : "") + ")\r\n";
            else
                result += "\r\n";

            result += "   Shared folder: " + share.directory.FullName + "\r\n";

            String auth = "";
            if (share.user != null && share.pass != null)
                auth = share.user + ":" + share.pass + "@";
            if (share.user != null && share.pass == null)
                auth = share.user + ":@";
            if (share.user == null && share.pass != null)
                auth = ":" + share.pass + "@";

            result += "   Host name: " + hostName +  "   FTP port: " + ftpPort + "   HTTP port: " + httpPort + "\r\n";
            if (hostIPv4 != null)
                result += "   IPv4 address: " + hostIPv4 + "\r\n";
            if (hostIPv6 != null)
                result += "   IPv6 address: " + hostIPv6 + "\r\n";

            result += "    (some handy URLs to access share)\r\n";

            String httpPortText = httpPort != 80 ? ":" + httpPort : "";
            result += "     http://" + hostName + httpPortText + share.rootDirName + "\r\n";
            if (hostIPv4 != null)
                result += "     http://" + hostIPv4 + httpPortText + share.rootDirName + "\r\n";
            if (hostIPv6 != null)
                result += "     http://[" + hostIPv6 + "]" + httpPortText + share.rootDirName + "\r\n";

            if (auth != "")
            {
                result += "     http://" + auth + hostName + httpPortText + share.rootDirName + "\r\n";
                if (hostIPv4 != null)
                    result += "     http://" + auth + hostIPv4 + httpPortText + share.rootDirName + "\r\n";
                if (hostIPv6 != null)
                    result += "     http://" + auth + "[" + hostIPv6 + "]" + httpPortText + share.rootDirName + "\r\n";
            }
            
            String ftpPortText = ftpPort != 21 ? ":" + ftpPort : "";
            result += "     ftp://" + hostName + ftpPortText + share.rootDirName + "/\r\n";
            if (hostIPv4 != null)
                result += "     ftp://" + hostIPv4 + ftpPortText + share.rootDirName + "/\r\n";
            if (hostIPv6 != null)
                result += "     ftp://[" + hostIPv6 + "]" + ftpPortText + share.rootDirName + "/\r\n";
            if (auth != "")
            {
                result += "     ftp://" + auth + hostName + ftpPortText + share.rootDirName + "/\r\n";
                if (hostIPv4 != null)
                    result += "     ftp://" + auth + hostIPv4 + ftpPortText + share.rootDirName + "/\r\n";
                if (hostIPv6 != null)
                    result += "     ftp://" + auth + "[" + hostIPv6 + "]" + ftpPortText + share.rootDirName + "/\r\n";
            }
            return result;
        }

        private static void WriteShare(FileStore.RootShare share, int ftpPort, int httpPort)
        {
            Write(GetShareInfo(share, ftpPort, httpPort));
        }

        [STAThread]
        static void Main(string[] args)
        {
            try
            {
                Control control = new Control();

                bool fromCommandPrompt = FromCommandPrompt();
                Attatch();
                if (args.Length == 0 && fromCommandPrompt)
                {
                    //display help info to command prompt
                    WriteHeader();
                    WriteHelp();
                    WriteFooter();
                }
                else if (args.Length == 1 && (!args[0].StartsWith("/") || args[0] == "/window"))
                {
                    Detatch();
                    Application.EnableVisualStyles();
                    Application.SetCompatibleTextRenderingDefault(false);
                    String path = null;
                    if (args[0] != "/window")
                    {
                        try
                        {
                            DirectoryInfo di = new DirectoryInfo(args[0]);
                            path = di.FullName;
                        }
                        catch (Exception) { }
                    }
                    Application.Run(new ManageForm(control, path));
                }
                else if (!fromCommandPrompt && args.Length == 0)
                {
                    Application.EnableVisualStyles();
                    Application.SetCompatibleTextRenderingDefault(false);
                    Application.Run(new ManageForm(control, null));
                }
                else
                {
                    if (fromCommandPrompt)
                        WriteHeader();
                    if (fromCommandPrompt && args[0] == "/help" || args[0] == "/?")
                    {
                        WriteHelp();
                    }
                    else if (fromCommandPrompt && args[0] == "/list")
                    {
                        WriteLine("Online Shares");
                        WriteLine("-------------");
                        List<FileStore.RootShare> shares = control.ListShares();
                        if (shares.Count == 0)
                            WriteLine(" (none)");
                        int[] ftpHttpPorts = null;
                        if (shares.Count != 0)
                            ftpHttpPorts = control.GetFtpHttpPorts();
                        foreach (FileStore.RootShare share in shares)
                        {
                            WriteShare(share, ftpHttpPorts[0], ftpHttpPorts[1]);
                        }
                    }
                    else if (args[0] == "/create")
                    {
                        if (args.Length < 4)
                        {
                            if (fromCommandPrompt)
                            {
                                WriteLine("Error: Please specify at least SHARENAME PATH PERMISSION");
                                FileInfo f = new FileInfo(Application.ExecutablePath);
                                WriteLine("To view options, type " + f.Name + " /help");
                            }
                        }
                        else
                        {
                            String url = args[1];
                            String path = args[2];
                            DirectoryInfo di = new DirectoryInfo(path);
                            path = di.FullName;
                            FileStore.Permission permission = FileStore.Permission.read;
                            if (args[3] == "rw")
                                permission = FileStore.Permission.readwrite;
                            String user = null;
                            if (args.Length >= 5)
                                user = args[4];
                            String pass = null;
                            if (args.Length >= 6)
                                pass = args[5];
                            FileStore.RootShare result = control.CreateShare(new FileStore.URL(url), path, permission, user, pass);
                            if (fromCommandPrompt)
                            {
                                if (result == null)
                                    WriteLine("Error: Failed creating share.");
                                else
                                {
                                    WriteLine("Share created:");
                                    int[] ftpHttpPorts = control.GetFtpHttpPorts();
                                    WriteShare(result, ftpHttpPorts[0], ftpHttpPorts[1]);
                                }
                            }
                        }
                    }
                    else if (args[0] == "/remove")
                    {
                        if (args.Length < 2)
                        {
                            if (fromCommandPrompt)
                            {
                                WriteLine("Error: Please specify at least SHARENAME");
                                FileInfo f = new FileInfo(Application.ExecutablePath);
                                WriteLine("To view options, type " + f.Name + " /help");
                            }
                        }
                        else
                        {
                            String url = args[1];
                            FileStore.URL shareURL = new FileStore.URL(url);
                            String result = control.DeleteShare(shareURL);
                            if (fromCommandPrompt)
                            {
                                if (result == "SUCCESS")
                                    WriteLine("Share successfully removed: " + shareURL.GetURL());
                                else
                                    WriteLine("Error: Failed removing share!");
                            }
                        }
                    }
                    else if (args[0] == "/removeall")
                    {
                        String result = control.DeleteAll();
                        if (fromCommandPrompt)
                        {
                            if (result == "SUCCESS")
                                WriteLine("All shares successfully removed!");
                            else
                                WriteLine("Error: Failed removing all shares!");
                        }
                    }
                    else
                    {
                        if (fromCommandPrompt)
                            WriteLine("Error: Unknown option!");
                    }
                    if (fromCommandPrompt)
                        WriteFooter();
                }
                Detatch();
                control.WaitForNoServer();
                if (update != null)
                {
                    if (!update.CheckIsDone())
                    {
                        //allow for 15 seconds to finish retrieving update info from web before terminating application
                        for (int i = 0; i < 15; i++)
                        {
                            Thread.Sleep(1000);
                            if (update.CheckIsDone())
                                break;
                        }
                    }
                }
            }
            finally
            {
                Log.Close();
                Application.Exit();
                Environment.Exit(0);
            }
            
        }


    }
}
